# Agilent_TwissTorr74 version 1.0 written by YS@KIT 2019

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


import qkit
from qkit.core.instrument_base import Instrument
import logging
import serial
import time
import atexit
import sys

class Agilent_TwissTorr74(Instrument):

    def __init__(self, name, address, reset=False):
        Instrument.__init__(self, name, tags=['physical'])
        logging.info(__name__ + ': Initializing instrument Twisstorr')
        if sys.version_info[0] < 3:
            def enc(string):
                return string
            def dec(string):
                return string
        else:
            def enc(string):
                return string.encode('latin-1')
            def dec(byte):
                return byte.decode('latin-1')
        self._enc = enc
        self._dec = dec
        self.setup(address)

    def setup(self, device="COM4"):
        # Open serial port, 9600, 8,N,1, timeout 1s
        baudrate = 9600
        timeout = 0.1

        # Serial port configuration
        ser = serial.Serial()
        ser.baudrate = baudrate
        ser.timeout = timeout
        self.ser = self._std_open(device, baudrate, timeout)
        atexit.register(self.ser.close)
        self._set_remote_serial(serial=True)
        if self.get_status == 0:
            self._set_active_stop(active_stop=True)

    def _std_open(self, device, baudrate, timeout):
        return serial.Serial(device, baudrate, timeout=timeout)

    def close(self):
        # Close serial port
        self.ser.close()

    def _remote_cmd(self, cmd):
        """
        Function sending the command string to the device.

        Args:
            cmd (str): Command string to be sent. Function "command" can be used to build it
                       from supplied window and data information (see pump manual for possible values)

        Returns:
            value (str): Answer from the device. Function "answer" can be use to unpack useful info.
        """
        self.ser.write(cmd)
        time.sleep(0.1)
        rem_char = self.ser.inWaiting()
        value = self.ser.read(rem_char)
        return value

    def command(self, addr=0x80, window='205', write=False, data=None):
        """
        Function building the command string sent to the device.
        For possible commands, see manual or use predefined functions in this module.

        Args:
            addr (hex): Unit address = 0x80 for RS-232, 0x80 + device number for RS-485
            window (str): String of 3 numeric character indicating the window number (defining the command)
            write (bool): Is the command used to read (False) or write (True) to the device
            data (str):  Alphanumeric ASCII string with the data to be written in case write==True

        Returns:
            cmd (str): Complete string to be sent to the device including a CRC
        """

        start = '\x02'
        command_string = start + chr(addr) + window

        if write:
            command_string += '1'
            command_string += data
        else:
            command_string += '0'

        end = '\x03'
        command_string += end

        crc = ord(command_string[1]) ^ ord(command_string[2])
        for i in range(3, len(command_string)):
            crc = crc ^ ord(command_string[i])
        crch = hex(crc)
        crc1 = crch[-2]
        crc2 = crch[-1]

        cmd = command_string + crc1 + crc2

        cmd = self._enc(cmd)

        return cmd

    def answer(self, answer):
        """
        Function unpacking the device answer.
        If the returned string has the length corresponding to one char of information,
        this information is decoded using the described meanings from the manual.
        In the case that the returned string is longer, the characters containing the data are returned.

        Args:
            answer (str): Answer string as it came from the device

        Returns:
            Decoded answer as described above.
        """
        answer = self._dec(answer)
        start = answer[0]
        address = answer[1]
        if len(answer) == 6:
            if answer[2] == chr(0x06):
                return answer[2], 'ACK'
            if answer[2] == chr(0x15):
                return answer[2], 'NACK'
            if answer[2] == chr(0x32):
                return answer[2], 'Unknown Window'
            if answer[2] == chr(0x33):
                return answer[2], 'Data Type Error'
            if answer[2] == chr(0x34):
                return answer[2], 'Out of Range'
            if answer[2] == chr(0x35):
                return answer[2], 'Window Read Only or Disabled'
            else:
                return answer
        else:
            window = answer[2] + answer[3] + answer[4]
            rw = answer[5]
            data = ''
            for i in range(6, len(answer) - 3):
                data += answer[i]
            return data

    def _set_remote_serial(self, serial=True):
        window = "008"
        write = True
        if serial:
            data = '0'
        else:
            data = '1'
        cmd = self.command(window=window, write=write, data=data)
        answer = self.answer(self._remote_cmd(cmd))
        return answer

    def _get_remote_serial(self):
        window = "008"
        write = False
        cmd = self.command(window=window, write=write)
        answer = self.answer(self._remote_cmd(cmd))
        return answer

    def _set_active_stop(self, active_stop=False):
        window = "107"
        write = True
        if active_stop:
            data = '1'
        else:
            data = '0'
        cmd = self.command(window=window, write=write, data=data)
        answer = self.answer(self._remote_cmd(cmd))
        return answer

    def _get_active_stop(self):
        window = "107"
        write = False
        cmd = self.command(window=window, write=write)
        answer = self.answer(self._remote_cmd(cmd))
        return int(answer)

    def get_status(self):
        """
        Get the pump status.

        Returns:
            status (int):
                0: 'Stop'
                1: 'Waiting Interlock'
                2: 'Starting'
                3: 'Auto-tuning'
                4: 'Breaking'
                5: 'Normal'
                6: 'Fail'
        """
        window = '205'
        write = False
        cmd = self.command(window=window, write=write)
        answer = self.answer(self._remote_cmd(cmd))
        meaning = ['Stop', 'Waiting Interlock', 'Starting', 'Auto-tuning', 'Braking', 'Normal', 'Fail']
        return int(answer), meaning[int(answer)]

    def get_error(self):
        """
        Get error code.

        Returns:
            Error code (Bitstring).
            Bit description as taken from manual: |7|6|5|4|3|2|1|0|
                7: Too high load
                6: Short circuit
                5: Over voltage
                4:
                3:
                2: Control over temperature
                1: Pump over temperature
                0: No connection
        """
        window = "206"
        write = False
        cmd = self.command(window=window, write=write)
        answer = self.answer(self._remote_cmd(cmd))
        return answer

    def get_rot_freq(self):
        """
        Get rotation frequency (Hz)

        Returns:
            freq (int): Rotation frequency in Hz
        """
        window = "226"
        write = False
        cmd = self.command(window=window, write=write)
        answer = self.answer(self._remote_cmd(cmd))
        answer = int(answer) / 60.
        return answer

    def get_temp(self):
        """
        Get pump temperature

        Returns:
            temp (int): Pump temperature in degC
        """
        window = '204'
        write = False
        cmd = self.command(window=window, write=write)
        answer = self.answer(self._remote_cmd(cmd))
        return int(answer)

    def get_power(self):
        """
        Get pump power

        Returns:
            power (int): Pump power in Watts
        """
        window = '202'
        write = False
        cmd = self.command(window=window, write=write)
        answer = self.answer(self._remote_cmd(cmd))
        return int(answer)

    def get_cycles(self):
        """
        Get pump cycle number

        Returns:
            cycle (int): Cycle number
        """
        window = '301'
        write = False
        cmd = self.command(window=window, write=write)
        answer = self.answer(self._remote_cmd(cmd))
        return int(answer)

    def get_pump_life(self):
        """
        Get pump life time

        Returns:
            pump life (int): Pump life in hours
        """
        window = '302'
        write = False
        cmd = self.command(window=window, write=write)
        answer = self.answer(self._remote_cmd(cmd))
        return int(answer)

    def start(self):
        """
        Start the pump
        """
        window = '000'
        write = True
        data = '1'
        cmd = self.command(window=window, write=write, data=data)
        answer = self.answer(self._remote_cmd(cmd))
        return answer

    def stop(self):
        """
        Stop the pump
        """
        window = "000"
        write = True
        data = '0'
        cmd = self.command(window=window, write=write, data=data)
        answer = self.answer(self._remote_cmd(cmd))
        return answer

    def set_softstart(self, status=False):
        """
        Set soft start option.
        Makes the pump start slower. Manual recommends after longer standing times.
        See manual.

        Args:
            status (bool): Status of soft start option to be set (standard: False)
        """
        window = "100"
        write = True
        if status:
            data = "1"
        else:
            data = "0"
        cmd = self.command(window=window, write=write, data=data)
        answer = self.answer(self._remote_cmd(cmd))
        return answer

    def get_softstart(self):
        """
        Get status of soft start option.
        Makes the pump start slower. Manual recommends after longer standing times.
        See manual.

        Returns:
            status (bool): Status of soft start option
        """
        window = "100"
        write = False
        cmd = self.command(window=window, write=write)
        answer = self.answer(self._remote_cmd(cmd))
        return int(answer)

    def set_rot_freq_setting(self, freq=1167):
        """
        Rotational frequency setting (Hz)

        Args:
            freq (int): Rotational frequency [1100..1167], default = 1167
        """
        window = "120"
        write = True
        data = str(freq)
        while len(data) < 6:
            data = '0' + data
        cmd = self.command(window=window, write=write, data=data)
        answer = self.answer(self._remote_cmd(cmd))
        return answer

    def get_rot_freq_setting(self):
        """
        Get rotational frequency setting (Hz)

        Returns:
            freq (int): Rotational frequency [1100..1167], default = 1167
        """
        window = "120"
        write = False
        cmd = self.command(window=window, write=write)
        answer = self.answer(self._remote_cmd(cmd))
        return int(answer)


# Tektronix_AWG5014.py class, to perform the communication between the Wrapper and the device
# Pieter de Groot <pieterdegroot@gmail.com>, 2008
# Martijn Schaafsma <qtlab@mcschaafsma.nl>, 2008
# Guenevere Prawiroatmodjo <guen@vvtp.tudelft.nl>, 2009
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from qkit.core.instrument_base import Instrument
from qkit import visa
import types
import logging
import numpy
import struct

class Tektronix_AWG5014(Instrument):
    '''
    This is the python driver for the Tektronix AWG5014
    Arbitrary Waveform Generator

    Usage:
    Initialize with
    <name> = instruments.create('name', 'Tektronix_AWG5014', address='<GPIB address>',
        reset=<bool>, numpoints=<int>)

    think about:    clock, waveform length

    TODO:
    1) Get All
    2) Remove test_send??
    3) Add docstrings
    4) Add 4-channel compatibility
    '''

    def __init__(self, name, address, reset=False, clock=1e9, numpoints=1000):
        '''
        Initializes the AWG520.

        Input:
            name (string)    : name of the instrument
            address (string) : GPIB address
            reset (bool)     : resets to default values, default=false
            numpoints (int)  : sets the number of datapoints

        Output:
            None
        '''
        logging.debug(__name__ + ' : Initializing instrument')
        Instrument.__init__(self, name, tags=['physical'])


        self._address = address
        self._visainstrument = visa.instrument(self._address)
        self._values = {}
        self._values['files'] = {}
        self._clock = clock
        self._numpoints = numpoints

        # Add parameters
        self.add_parameter('waveform', type=str,
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            channels=(1, 4), channel_prefix='ch%d_')
        self.add_parameter('output', type=float,
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            channels=(1, 4), channel_prefix='ch%d_')
        self.add_parameter('wlist', type=str,
            flags=Instrument.FLAG_GET)
        self.add_parameter('trigger_mode', type=str,
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET)
        self.add_parameter('trigger_impedance', type=float,
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            minval=49, maxval=2e3, units='Ohm')
        self.add_parameter('trigger_level', type=float,
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            minval=-5, maxval=5, units='Volts')
        self.add_parameter('clock', type=float,
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            minval=1e6, maxval=1e9, units='Hz')
        self.add_parameter('numpoints', type=int,
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            minval=100, maxval=1e9, units='Int')
        self.add_parameter('filename', type=str,
            flags=Instrument.FLAG_SET, channels=(1, 4),
            channel_prefix='ch%d_')
        self.add_parameter('amplitude', type=float,
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            channels=(1, 4), minval=0, maxval=2, units='Volts', channel_prefix='ch%d_')
        self.add_parameter('offset', type=float,
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            channels=(1, 4), minval=-2, maxval=2, units='Volts', channel_prefix='ch%d_')
        self.add_parameter('marker1_low', type=float,
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            channels=(1, 4), minval=-2, maxval=2, units='Volts', channel_prefix='ch%d_')
        self.add_parameter('marker1_high', type=float,
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            channels=(1, 4), minval=-2, maxval=2, units='Volts', channel_prefix='ch%d_')
        self.add_parameter('marker2_low', type=float,
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            channels=(1, 4), minval=-2, maxval=2, units='Volts', channel_prefix='ch%d_')
        self.add_parameter('marker2_high', type=float,
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            channels=(1, 4), minval=-2, maxval=2, units='Volts', channel_prefix='ch%d_')
        self.add_parameter('status', type=bool,
            flags=Instrument.FLAG_GETSET | Instrument.FLAG_GET_AFTER_SET,
            channels=(1, 4),channel_prefix='ch%d_')

        # Add functions
        self.add_function('reset')
        self.add_function('get_all')
        self.add_function('clear_waveforms')
        self.add_function('set_trigger_mode_on')
        self.add_function('set_trigger_mode_off')
        self.add_function('set_trigger_impedance_1e3')
        self.add_function('set_trigger_impedance_50')

        # Make Load/Delete Waveform functions for each channel
        for ch in range(1,5):
            self._add_load_waveform_func(ch)
            self._add_del_loaded_waveform_func(ch)

        if reset:
            self.reset()
        else:
            self.get_all()

    # Functions
    def reset(self):
        '''
        Resets the instrument to default values

        Input:
            None

        Output:
            None
        '''
        logging.info(__name__ + ' : Resetting instrument')
        self._visainstrument.write('*RST')

    def get_all(self):
        '''
        Reads all implemented parameters from the instrument,
        and updates the wrapper.

        Input:
            None

        Output:
            None
        '''
        logging.info(__name__ + ' : Reading all data from instrument')
        logging.warning(__name__ + ' : get all not yet fully functional')

        self.get_trigger_mode()
        self.get_trigger_impedance()
        self.get_trigger_level()
        self.get_numpoints()
        self.get_clock()

        for i in range(1,3):
            self.get('ch%d_amplitude' % i)
            self.get('ch%d_offset' % i)
            self.get('ch%d_marker1_low' % i)
            self.get('ch%d_marker1_high' % i)
            self.get('ch%d_marker2_low' % i)
            self.get('ch%d_marker2_high' % i)
            self.get('ch%d_status' % i)

    def clear_waveforms(self):
        '''
        Clears the waveform on both channels.

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Clear waveforms from channels')
        self._visainstrument.write('SOUR1:FUNC:USER ""')
        self._visainstrument.write('SOUR2:FUNC:USER ""')
        self._visainstrument.write('SOUR3:FUNC:USER ""')
        self._visainstrument.write('SOUR4:FUNC:USER ""')

    def run(self):
        '''
        Initiates the output of a waveform or a sequence. This is equivalent to pressing
        Run/Delete/Stop button on the front panel. The instrument can be put in the run
        state only when output waveforms are assigned to channels.

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Run/Initiate output of a waveform or sequence')
        self._visainstrument.write('AWGC:RUN:IMM')

    def stop(self):
        '''
        Terminates the output of a waveform or a sequence. This is equivalent to pressing
        Run/Delete/Stop button on the front panel.

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Stop/Terminate output of a waveform or sequence')
        self._visainstrument.write('AWGC:STOP:IMM')

    def do_set_output(self, state, channel):
        '''
        This command sets the output state of the AWG.
        Input:
            channel (int) : the source channel
            state (int) : on (1) or off (0)

        Output:
            None
        '''
        logging.debug(__name__ + ' : Set channel output state')
        if (state == 1):
            self._visainstrument.write('OUTP%s:STAT ON' % channel)
        if (state == 0):
            self._visainstrument.write('OUTP%s:STAT OFF' % channel)

    def do_get_output(self, channel):
        '''
        This command gets the output state of the AWG.
        Input:
            channel (int) : the source channel

        Output:
            state (int) : on (1) or off (0)
        '''
        logging.debug(__name__ + ' : Get channel output state')
        return self._visainstrument.ask('OUTP%s:STAT?' % channel)

    def do_set_waveform(self, waveform, channel):
        '''
        This command sets the output waveform from the current waveform
        list for each channel when Run Mode is not Sequence.

        Input:
            channel (int) : the source channel
            waveform (str) : the waveform filename as loaded in waveform list

        Output:
            None
        '''
        logging.debug(__name__ + ' : Set the output waveform for channel %s' % channel)
        self._visainstrument.write('SOUR%s:WAV "%s"' % (channel, waveform))

    def do_get_waveform(self, channel):
        '''
        This command returns the output waveform from the current waveform
        list for each channel when Run Mode is not Sequence.

        Input:
            channel (int) : the source channel

        Output:
            waveform (str) : the waveform filename as loaded in waveform list
        '''
        logging.debug(__name__ + ' : Get the output waveform for channel %s' % channel)
        return self._visainstrument.ask('SOUR%s:WAV?' % channel)

    def do_get_wlist(self):
        '''
        This command returns the waveform list in an array.
        Input:
            None

        Output:
            wlist (array) : the waveform list in an array.
        '''
        size = int(self._visainstrument.ask('WLIST:SIZE?'))
        wlist = []
        for i in range(0, size):
            wname = self._visainstrument.ask('WLIST:NAME? %f' % i)
            wname = wname.replace('"','')
            wlist.append(wname)
        return wlist

    def del_waveform(self, name):
        '''
        This command deletes the waveform from the waveform list.
        Input:
            name (str) : waveform name, as defined in the waveform list

        Output:
            None
        '''
        logging.debug(__name__ + ' : Delete the waveform "%s" from the waveform list' % name)
        self._visainstrument.write('WLIS:WAV:DEL "%s"' % name)

    def del_loaded_waveform(self, channel):
        '''
        This command deletes the waveform from the waveform list which was loaded
        on a channel.
        Input:
            name (str) : waveform name, as defined in the waveform list
            channel (int) : channel (1,4)

        Output:
            None
        '''
        name = 'CH%sWFM' % channel
        self.del_waveform(name)

    def del_waveform_all(self):
        '''
        This command deletes all waveforms in the user-defined waveform list.
        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__ + ' : Clear waveform list')
        self._visainstrument.write('WLIS:WAV:DEL ALL')

    def load_waveform(self, channel, filename, drive='Z:', path='\\'):
        '''
        Use this command to directly load a sequence file or a waveform file to a specific channel.

        Input:
            channel (int) : the source channel
            filename (str) : the waveform filename (.wfm, .seq)
            drive (str) : the local drive where the file is located (e.g. 'Z:')
            path (str) : the local path where the file is located (e.g. '\waveforms')

        Output:
            None
        '''
        logging.debug(__name__ + ' : Load waveform file %s%s%s for channel %s' % (drive, path, filename, channel))
        self._visainstrument.write('SOUR%s:FUNC:USER "%s/%s","%s"' % (channel, path, filename, drive))

    def _add_load_waveform_func(self, channel):
        '''
        Adds load_ch[n]_waveform functions, based on load_waveform(channel, filename, drive, path).
        n = (1,2,3,4) for 4 channels.
        '''
        func = lambda filename, drive='Z:', path='\\': self.load_waveform(channel, filename, drive, path)
        setattr(self, 'load_ch%s_waveform' % channel, func)

    def _add_del_loaded_waveform_func(self, channel):
        '''
        Adds load_ch[n]_waveform functions, based on load_waveform(channel, filename, drive, path).
        n = (1,2,3,4) for 4 channels.
        '''
        func = lambda: self.del_loaded_waveform(channel)
        setattr(self, 'del_ch%s_waveform' % channel, func)

    def load_settings(self, filename, drive='Z:', path='\\'):
        '''
        This command sets the AWG's setting from the specified settings file.

        Input:
            filename (str) : the settings filename (.set)
            drive (str) : the settings file drive
            path (str) : the settings file path

        Output:
            None
        '''
        logging.debug(__name__ + ' : Load settings file %s%s%s' % (drive, path, filename))
        self._visainstrument.write('AWGC:SRES "%s","%s"' % (filename, drive))

    def save_settings(self, filename, drive='Z:', path='\\'):
        '''
        This command saves the AWG's current setting to the specified settings file.
        Default path is the Z:\ drive, , which is located at
        "C:\Documents and Settings\All Users\Documents\Waveforms".

        Input:
            filename (str) : the settings file path (.set)
            drive (str) : the settings file drive
            path (str) : the settings file path

        Output:
            None
        '''
        logging.debug(__name__ + ' : Save current settings to file %s' % filename)
        self._visainstrument.write('AWGC:SSAV "%s","%s"' % (filename, drive))

    def set_runmode(self, runmode):
        '''
        Set the Run Mode of the device to Continuous, Triggered, Gated or Sequence.
        Input:
            runmode (str) : The Run mode which can be set to 'CONT', 'TRIG', 'GAT' or 'SEQ'.

        Output:
            None
        '''
        logging.debug(__name__ + ' : Set runmode to %s' % runmode)
        runmode = runmode.upper()
        if ((runmode == 'TRIG') | (runmode == 'CONT') | (runmode == 'SEQ') | (runmode == 'GATE')):
            self._visainstrument.write('AWGC:RMOD %s' % runmode)
        else:
            logging.error(__name__ + ' : Unable to set trigger mode to %s, expected "CONT", "TRIG", "GATE" or "SEQ"' % runmode)

    def set_sequence_mode_on(self):
        '''
        Sets the sequence mode to 'On'

        Input:
            None

        Output:
            None
        '''
        self.set_runmode('SEQ')

    def set_trigger_mode_on(self):
        '''
        Sets the trigger mode to 'On'

        Input:
            None

        Output:
            None
        '''
        self.set_runmode('TRIG')

    def set_runmode_cont(self):
        '''
        Sets the trigger mode to 'Cont'

        Input:
            None

        Output:
            None
        '''
        self.set_runmode('CONT')

    def set_trigger_impedance_1e3(self):
        '''
        Sets the trigger impedance to 1 kOhm

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__  + ' : Set trigger impedance to 1e3 Ohm')
        self._visainstrument.write('TRIG:IMP 1e3')

    def set_trigger_impedance_50(self):
        '''
        Sets the trigger impedance to 50 Ohm

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__  + ' : Set trigger impedance to 50 Ohm')
        self._visainstrument.write('TRIG:IMP 50')

    # Parameters
    def do_get_trigger_mode(self):
        '''
        Reads the trigger mode from the instrument

        Input:
            None

        Output:
            mode (string) : 'Trig' or 'Cont' depending on the mode
        '''
        logging.debug(__name__  + ' : Get trigger mode from instrument')
        return self._visainstrument.ask('AWGC:RMOD?')

    def do_set_trigger_mode(self, mod):
        '''
        Sets trigger mode of the instrument

        Input:
            mod (string) : Either 'Trig' or 'Cont' depending on the mode

        Output:
            None
        '''
        if (mod.upper()=='TRIG'):
            self.set_trigger_mode_on()
        elif (mod.upper()=='CONT'):
            self.set_runmode('CONT')
        else:
            logging.error(__name__ + ' : Unable to set trigger mode to %s, expected "TRIG" or "CONT"' % mod)

    def do_get_trigger_impedance(self):
        '''
        Reads the trigger impedance from the instrument

        Input:
            None

        Output:
            impedance (??) : 1e3 or 50 depending on the mode
        '''
        logging.debug(__name__  + ' : Get trigger impedance from instrument')
        return self._visainstrument.ask('TRIG:IMP?')

    def do_set_trigger_impedance(self, mod):
        '''
        Sets the trigger impedance of the instrument

        Input:
            mod (int) : Either 1e3 of 50 depending on the mode

        Output:
            None
        '''
        if (mod==1e3):
            self.set_trigger_impedance_1e3()
        elif (mod==50):
            self.set_trigger_impedance_50()
        else:
            logging.error(__name__ + ' : Unable to set trigger impedance to %s, expected "1e3" or "50"' % mod)

    def do_get_trigger_level(self):
        '''
        Reads the trigger level from the instrument

        Input:
            None

        Output:
            None
        '''
        logging.debug(__name__  + ' : Get trigger level from instrument')
        return float(self._visainstrument.ask('TRIG:LEV?'))

    def do_set_trigger_level(self, level):
        '''
        Sets the trigger level of the instrument

        Input:
            level (float) : trigger level in volts
        '''
        logging.debug(__name__  + ' : Trigger level set to %.3f' % level)
        self._visainstrument.write('TRIG:LEV %.3f' % level)

    def do_get_numpoints(self):
        '''
        Returns the number of datapoints in each wave

        Input:
            None

        Output:
            numpoints (int) : Number of datapoints in each wave
        '''
        return self._numpoints

    def do_set_numpoints(self, numpts):
        '''
        Sets the number of datapoints in each wave.
        This acts on all channels.

        Input:
            numpts (int) : The number of datapoints in each wave

        Output:
            None
        '''
        logging.debug(__name__ + ' : Trying to set numpoints to %s' % numpts)
        if numpts != self._numpoints:
            logging.warning(__name__ + ' : changing numpoints. This will clear all waveforms!')

        response = raw_input('type "yes" to continue')
        if response is 'yes':
            logging.debug(__name__ + ' : Setting numpoints to %s' % numpts)
            self._numpoints = numpts
            self.clear_waveforms()
        else:
            print 'aborted'

    def do_get_clock(self):
        '''
        Returns the clockfrequency, which is the rate at which the datapoints are
        sent to the designated output

        Input:
            None

        Output:
            clock (int) : frequency in Hz
        '''
        return self._clock

    def do_set_clock(self, clock):
        '''
        Sets the rate at which the datapoints are sent to the designated output channel

        Input:
            clock (int) : frequency in Hz

        Output:
            None
        '''
        logging.warning(__name__ + ' : Clock set to %s. This is not fully functional yet. To avoid problems, it is better not to change the clock during operation' % clock)
        self._clock = clock
        self._visainstrument.write('SOUR:FREQ %f' % clock)

    def do_set_filename(self, name, channel):
        '''
        Specifies which file has to be set on which channel
        Make sure the file exists, and the numpoints and clock of the file
        matches the instrument settings.

        If file doesn't exist an error is raised, if the numpoints doesn't match
        the command is neglected

        Input:
            name (string) : filename of uploaded file
            channel (int) : 1 or 2, the number of the designated channel

        Output:
            None
        '''
        logging.debug(__name__  + ' : Try to set %s on channel %s' % (name, channel))
        exists = False
        if name in self._values['files']:
            exists= True
            logging.debug(__name__  + ' : File exists in local memory')
            self._values['recent_channel_%s' % channel] = self._values['files'][name]
            self._values['recent_channel_%s' % channel]['filename'] = name
        else:
            logging.debug(__name__  + ' : File does not exist in memory, \
            reading from instrument')
            lijst = self._visainstrument.ask('MMEM:CAT? "MAIN"')
            bool = False
            bestand=""
            for i in range(len(lijst)):
                if (lijst[i]=='"'):
                    bool=True
                elif (lijst[i]==','):
                    bool=False
                    if (bestand==name): exists=True
                    bestand=""
                elif bool:
                    bestand = bestand + lijst[i]
        if exists:
            data = self._visainstrument.ask('MMEM:DATA? "%s"' % name)
            logging.debug(__name__  + ' : File exists on instrument, loading \
            into local memory')
            # string alsvolgt opgebouwd: '#' <lenlen1> <len> 'MAGIC 1000\r\n' '#' <len waveform> 'CLOCK ' <clockvalue>
            len1=int(data[1])
            len2=int(data[2:2+len1])
            i=len1
            tekst = ""
            while (tekst!='#'):
                tekst=data[i]
                i=i+1
            len3=int(data[i])
            len4=int(data[i+1:i+1+len3])

            w=[]
            m1=[]
            m2=[]

            for q in range(i+1+len3, i+1+len3+len4,5):
                j=int(q)
                c,d = struct.unpack('<fB', data[j:5+j])
                w.append(c)
                m2.append(int(d/2))
                m1.append(d-2*int(d/2))

            clock = float(data[i+1+len3+len4+5:len(data)])

            self._values['files'][name]={}
            self._values['files'][name]['w']=w
            self._values['files'][name]['m1']=m1
            self._values['files'][name]['m2']=m2
            self._values['files'][name]['clock']=clock
            self._values['files'][name]['numpoints']=len(w)

            self._values['recent_channel_%s' % channel] = self._values['files'][name]
            self._values['recent_channel_%s' % channel]['filename'] = name
        else:
            logging.error(__name__  + ' : Invalid filename specified %s' % name)

        if (self._numpoints==self._values['files'][name]['numpoints']):
            logging.debug(__name__  + ' : Set file %s on channel %s' % (name, channel))
            self._visainstrument.write('SOUR%s:FUNC:USER "%s","MAIN"' % (channel, name))
        else:
            logging.warning(__name__  + ' : Verkeerde lengte %s ipv %s'
                % (self._values['files'][name]['numpoints'], self._numpoints))

    def do_get_amplitude(self, channel):
        '''
        Reads the amplitude of the designated channel from the instrument

        Input:
            channel (int) : 1 or 2, the number of the designated channel

        Output:
            amplitude (float) : the amplitude of the signal in Volts
        '''
        logging.debug(__name__ + ' : Get amplitude of channel %s from instrument'
            % channel)
        return float(self._visainstrument.ask('SOUR%s:VOLT:LEV:IMM:AMPL?' % channel))

    def do_set_amplitude(self, amp, channel):
        '''
        Sets the amplitude of the designated channel of the instrument

        Input:
            amp (float)   : amplitude in Volts
            channel (int) : 1 or 2, the number of the designated channel

        Output:
            None
        '''
        logging.debug(__name__ + ' : Set amplitude of channel %s to %.6f'
            % (channel, amp))
        self._visainstrument.write('SOUR%s:VOLT:LEV:IMM:AMPL %.6f' % (channel, amp))

    def do_get_offset(self, channel):
        '''
        Reads the offset of the designated channel of the instrument

        Input:
            channel (int) : 1 or 2, the number of the designated channel

        Output:
            offset (float) : offset of designated channel in Volts
        '''
        logging.debug(__name__ + ' : Get offset of channel %s' % channel)
        return float(self._visainstrument.ask('SOUR%s:VOLT:LEV:IMM:OFFS?' % channel))

    def do_set_offset(self, offset, channel):
        '''
        Sets the offset of the designated channel of the instrument

        Input:
            offset (float) : offset in Volts
            channel (int)  : 1 or 2, the number of the designated channel

        Output:
            None
        '''
        logging.debug(__name__ + ' : Set offset of channel %s to %.6f' % (channel, offset))
        self._visainstrument.write('SOUR%s:VOLT:LEV:IMM:OFFS %.6f' % (channel, offset))

    def do_get_marker1_low(self, channel):
        '''
        Gets the low level for marker1 on the designated channel.

        Input:
            channel (int) : 1 or 2, the number of the designated channel

        Output:
            low (float) : low level in Volts
        '''
        logging.debug(__name__ + ' : Get lower bound of marker1 of channel %s' % channel)
        return float(self._visainstrument.ask('SOUR%s:MARK1:VOLT:LEV:IMM:LOW?' % channel))

    def do_set_marker1_low(self, low, channel):
        '''
        Sets the low level for marker1 on the designated channel.

        Input:
            low (float)   : low level in Volts
            channel (int) : 1 or 2, the number of the designated channel

        Output:
            None
         '''
        logging.debug(__name__ + ' : Set lower bound of marker1 of channel %s to %.3f'
            % (channel, low))
        self._visainstrument.write('SOUR%s:MARK1:VOLT:LEV:IMM:LOW %.3f' % (channel, low))

    def do_get_marker1_high(self, channel):
        '''
        Gets the high level for marker1 on the designated channel.

        Input:
            channel (int) : 1 or 2, the number of the designated channel

        Output:
            high (float) : high level in Volts
        '''
        logging.debug(__name__ + ' : Get upper bound of marker1 of channel %s' % channel)
        return float(self._visainstrument.ask('SOUR%s:MARK1:VOLT:LEV:IMM:HIGH?' % channel))

    def do_set_marker1_high(self, high, channel):
        '''
        Sets the high level for marker1 on the designated channel.

        Input:
            high (float)   : high level in Volts
            channel (int) : 1 or 2, the number of the designated channel

        Output:
            None
         '''
        logging.debug(__name__ + ' : Set upper bound of marker1 of channel %s to %.3f'
            % (channel, high))
        self._visainstrument.write('SOUR%s:MARK1:VOLT:LEV:IMM:HIGH %.3f' % (channel, high))

    def do_get_marker2_low(self, channel):
        '''
        Gets the low level for marker2 on the designated channel.

        Input:
            channel (int) : 1 or 2, the number of the designated channel

        Output:
            low (float) : low level in Volts
        '''
        logging.debug(__name__ + ' : Get lower bound of marker2 of channel %s' % channel)
        return float(self._visainstrument.ask('SOUR%s:MARK2:VOLT:LEV:IMM:LOW?' % channel))

    def do_set_marker2_low(self, low, channel):
        '''
        Sets the low level for marker2 on the designated channel.

        Input:
            low (float)   : low level in Volts
            channel (int) : 1 or 2, the number of the designated channel

        Output:
            None
         '''
        logging.debug(__name__ + ' : Set lower bound of marker2 of channel %s to %.3f'
            % (channel, low))
        self._visainstrument.write('SOUR%s:MARK2:VOLT:LEV:IMM:LOW %.3f' % (channel, low))

    def do_get_marker2_high(self, channel):
        '''
        Gets the high level for marker2 on the designated channel.

        Input:
            channel (int) : 1 or 2, the number of the designated channel

        Output:
            high (float) : high level in Volts
        '''
        logging.debug(__name__ + ' : Get upper bound of marker2 of channel %s' % channel)
        return float(self._visainstrument.ask('SOUR%s:MARK2:VOLT:LEV:IMM:HIGH?' % channel))

    def do_set_marker2_high(self, high, channel):
        '''
        Sets the high level for marker2 on the designated channel.

        Input:
            high (float)   : high level in Volts
            channel (int) : 1 or 2, the number of the designated channel

        Output:
            None
         '''
        logging.debug(__name__ + ' : Set upper bound of marker2 of channel %s to %.3f'
            % (channel, high))
        self._visainstrument.write('SOUR%s:MARK2:VOLT:LEV:IMM:HIGH %.3f' % (channel, high))

    def do_get_status(self, channel):
        '''
        Gets the status of the designated channel.

        Input:
            channel (int) : 1 or 2, the number of the designated channel

        Output:
            None
        '''
        logging.debug(__name__ + ' : Get status of channel %s' % channel)
        outp = self._visainstrument.ask('OUTP%s?' % channel)
        try:
            return bool(outp)
        except:
            logging.debug(__name__ + ' : Read invalid status from instrument %s' % outp)
            return 'an error occurred while reading status from instrument'

    def do_set_status(self, status, channel):
        '''
        Sets the status of designated channel.

        Input:
            status (boolean) 
            channel (int)   : channel number

        Output:
            None
        '''
        logging.debug(__name__ + ' : Set status of channel %s to %s'
            % (channel, status))
        if status:
            self._visainstrument.write('OUTP%s ON' % channel)
        elif not status:
            self._visainstrument.write('OUTP%s OFF' % channel)
        else:
            logging.debug(__name__ + ' : Try to set status to invalid value %s' % status)
            print 'Tried to set status to invalid value %s' % status

    #  Ask for string with filenames
    def get_filenames(self):
        logging.debug(__name__ + ' : Read filenames from instrument')
        return self._visainstrument.ask('MMEM:CAT? "MAIN"')

    # Send waveform to the device
    def send_waveform(self,w,m1,m2,filename,clock):
        '''
        Sends a complete waveform. All parameters need to be specified.
        See also: resend_waveform()

        Input:
            w (float[numpoints]) : waveform
            m1 (int[numpoints])  : marker1
            m2 (int[numpoints])  : marker2
            filename (string)    : filename
            clock (int)          : frequency (Hz)

        Output:
            None
        '''
        logging.debug(__name__ + ' : Sending waveform %s to instrument' % filename)
        # Check for errors
        dim = len(w)

        if (not((len(w)==len(m1)) and ((len(m1)==len(m2))))):
            return 'error'

        self._values['files'][filename]={}
        self._values['files'][filename]['w']=w
        self._values['files'][filename]['m1']=m1
        self._values['files'][filename]['m2']=m2
        self._values['files'][filename]['clock']=clock
        self._values['files'][filename]['numpoints']=len(w)

        m = m1 + numpy.multiply(m2,2)
        ws = ''
        for i in range(0,len(w)):
            ws = ws + struct.pack('<fB', w[i], int(m[i]))

        s1 = 'MMEM:DATA "%s",' % filename
        s3 = 'MAGIC 1000\n'
        s5 = ws
        s6 = 'CLOCK %.10e\n' % clock

        s4 = '#' + str(len(str(len(s5)))) + str(len(s5))
        lenlen=str(len(str(len(s6) + len(s5) + len(s4) + len(s3))))
        s2 = '#' + lenlen + str(len(s6) + len(s5) + len(s4) + len(s3))

        mes = s1 + s2 + s3 + s4 + s5 + s6

        self._visainstrument.write(mes)

    def resend_waveform(self, channel, w=[], m1=[], m2=[], clock=[]):
        '''
        Resends the last sent waveform for the designated channel
        Overwrites only the parameters specified

        Input: (mandatory)
            channel (int) : 1, 2, 3 or 4, the number of the designated channel

        Input: (optional)
            w (float[numpoints]) : waveform
            m1 (int[numpoints])  : marker1
            m2 (int[numpoints])  : marker2
            clock (int) : frequency

        Output:
            None
        '''
        filename = self._values['recent_channel_%s' % channel]['filename']
        logging.debug(__name__ + ' : Resending %s to channel %s' % (filename, channel))


        if (w==[]):
            w = self._values['recent_channel_%s' % channel]['w']
        if (m1==[]):
            m1 = self._values['recent_channel_%s' % channel]['m1']
        if (m2==[]):
            m2 = self._values['recent_channel_%s' % channel]['m2']
        if (clock==[]):
            clock = self._values['recent_channel_%s' % channel]['clock']

        if not ( (len(w) == self._numpoints) and (len(m1) == self._numpoints) and (len(m2) == self._numpoints)):
            logging.error(__name__ + ' : one (or more) lengths of waveforms do not match with numpoints')

        self.send_waveform(w,m1,m2,filename,clock)
        self.do_set_filename(filename, channel)




# Nitrogen refilling automation by JB@KIT 11/2015 jochen.braumueller@kit.edu
# model: XXX + custom controller box to monitor/switch off nitrogen refilling (R. Jehle)

# this file is named 'NitrogenControl.py' and runs on the raspberry in the controller box
# functionality: logging of switching events to 'nitrogen.log', controlled stop of refilling when
#   automatic refilling continues for more than 30? minutes

# model: Raspberry Pi running Raspian
# server scripts has to be run with su permission due to GPIO usage

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# install:
'''
    sudo apt-get update
    sudo apt-get install python-dev
    sudo apt-get install python-rpi.gpio
'''

# GPIO settings
import RPi.GPIO as GPIO
GPIO.setmode(GPIO.BOARD)   #use RPi.GPIO layout

OUTPUT = 21   #gpio9
SENSE = 23   #gpio11
SWITCH = 19   #gpio10
GPIO.setup(OUTPUT, GPIO.OUT)
GPIO.setup(SENSE, GPIO.IN)
GPIO.output(OUTPUT, GPIO.LOW)   #switch on by default THIS RELAY IS INVERSE!
GPIO.setup(SWITCH, GPIO.OUT)
GPIO.output(SWITCH, GPIO.LOW)   #manual switch is off per default

try:
    import SocketServer
except ImportError:
    import socketserver as SocketServer
from threading import Thread
from sendMail import *

#import os,sys
from time import sleep


def _send_mail(m = ''):
    
    strTime = time.strftime('%x %X')

    s = SendMail("UFO@kit.edu",strTime+" UFO monitoring - manual switch off of nitrogen refilling")   #sender, subject
    
    s.setRecipients("")
    #s.add_Line("Hello, I am feeling great. Cool gases. Nice pressures.")
    s.add_Line(strTime)
    s.add_Line("The control script monitoring nitrogen refilling performed a manual switch off of nitrogen refilling.")
    s.add_Line("Refilling was taking for more than 60 minutes.")
    s.add_Line("A possible reason for this is a nitrogen dewar running empty or a malfunction of the sensor in the nitrogen bath or the PID controller.")
    s.add_Line("")
    s.add_Line(" - What happened last night can happen again. - ")
    s.add_Line("")
    s.add_Line("Please reflect this event critically. Refilling can be turned on again via remote by running the script refilling_switch_on.py, typing")
    s.add_Line("python /home/pi/nitrogen/refilling_switch_on.py")
    s.add_Line("in a ssh shell connecting to pi-us87.")
    s.add_Line("")
    if m != '':
        s.add_Line("Additional message: " + str(m))
        s.add_Line("")
    s.add_Line("Best, UFO")
    
    s.sendEMail()
    print "Mail sent to recipients."
    

class TCPHandler_Nitrogen(SocketServer.BaseRequestHandler):
    """
    The RequestHandler class for our server.

    It is instantiated once per connection to the server, and must
    override the handle() method to implement communication to the
    client.
    """
    
    def set_output(self, switch_to):
        if switch_to == 'on':
            GPIO.output(OUTPUT, GPIO.LOW)
            self.man_switch = True
            sleep(0.1)
            return True
        elif switch_to == 'off':
            GPIO.output(OUTPUT, GPIO.HIGH)
            self.man_switch = False
            sleep(0.1)
            _send_mail()
            return True
        else:
            return False
    def get_state(self):
        #for i in range(15): #As we have an AC component on the signal, we have to average...
        if GPIO.input(SENSE) == GPIO.HIGH:
            return True
            #sleep(.1)
        return False

            
    def handle(self):
        try:
            # self.request is the TCP socket connected to the client
            self.data = self.request.recv(1024).strip()
            if str(self.data) == 'get_status':
                print 'status request from', str(format(self.client_address[0]))
                if self.get_state():   #refilling active
                    status = '1'
                else:
                    status = '0'
                self.request.sendall(str(status))
            elif str(self.data) == 'output on':
                print 'output on request from', str(format(self.client_address[0]))
                self.request.sendall(str(self.set_output('on')))
            elif str(self.data) == 'output off':
                print 'output off request from', str(format(self.client_address[0]))
                self.request.sendall(str(self.set_output('off')))
            else:
                print 'request string not recognized'
                self.request.sendall(0)
        except Exception as m:
            print 'Error in TCP server handler:', m


class ThreadedTCPServer(SocketServer.ThreadingMixIn, SocketServer.TCPServer):   #enable multiple access
    pass


if __name__ == "__main__":
    HOST, PORT = 'ip-address', 9989

    # Create the server, binding to localhost on port 9955
    server = ThreadedTCPServer((HOST, PORT), TCPHandler_Nitrogen)

    # Activate the server; this will keep running until you
    # interrupt the program with Ctrl-C
    print 'Starting server...'
    
    try:
        server.serve_forever()
    finally:
        print 'shut down server'
        server.shutdown()




# -*- coding: utf-8 -*-

# Oxford_ITC503.py driver for Oxford ITC503 temperature controller connected 
# via a Prologix GPIB ethernet controller
# Micha Wildermuth, micha.wildermuth@kit.edu 2019
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from qkit.core.instrument_base import Instrument
from qkit import visa
import re
import logging
        

class Oxford_ITC503(Instrument):
    """
    This is the driver for the Oxford ITC503 temperature controller
    """
    
    def __init__(self, name, address):
        """
        Initializes VISA communication with the instrument Keithley 2636A.
        
        Parameters
        ----------
        name: string
            Name of the instrument (driver).
        address: string
            IP address for the communication with the instrument.
        
        Returns
        -------
        None
        
        Examples
        --------
        
        >>> import qkit
        QKIT configuration initialized -> available as qkit.cfg[...]
        
        >>> qkit.start()
        Starting QKIT framework ... -> qkit.core.startup
        Loading module ... S10_logging.py
        Loading module ... S12_lockfile.py
        Loading module ... S14_setup_directories.py
        Loading module ... S20_check_for_updates.py
        Loading module ... S25_info_service.py
        Loading module ... S30_qkit_start.py
        Loading module ... S65_load_RI_service.py
        Loading module ... S70_load_visa.py
        Loading module ... S80_load_file_service.py
        Loading module ... S85_init_measurement.py
        Loading module ... S98_started.py
        Loading module ... S99_init_user.py
        
        >>> ITC503 = qkit.instruments.create('ITC503', 'Oxford_ITC503', address='ASRL1::INSTR')
        """
        self.__name__ = __name__
        # Start VISA communication
        logging.info(__name__ + ': Initializing instrument Oxford ITC 503 temperature control')
        Instrument.__init__(self, name, tags=['physical'])
        self._address = address
        self._visainstrument = visa.instrument(self._address)
        self._visainstrument.read_termination = '\r'
    
    def write(self, cmd):
        """
        Sends a visa command <cmd> to the Device.
        
        Parameters
        ----------
        cmd: str
            Command that is send to the instrument via pyvisa and NI-VISA backend.
        
        Returns
        -------
        None
        """
        #if cmd == '*CLS': cmd = 'Q0'
        ans = self._visainstrument.query(cmd)  # Use query, since it returns first letter of <cmd>, e.g. u'O'
        return ans
    
    def query(self, cmd):
        """
        Clears the communication buffer, sends a visa command <cmd> and returns the read answer <ans>.
        
        Parameters
        ----------
        cmd: str
            Command that is send to the instrument via pyvisa and NI-VISA backend.
        
        Returns
        -------
        ans: str
            Answer that is returned at query after the sent <cmd>.
        """
        self._visainstrument.write('Q0')  # usually defines the communication protocol, but can be used to clear the the communications buffer
        return self._visainstrument.query(cmd)
    
    def get_IDN(self):
        """
        Gets the identification query of the device.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        IDN: (str)
            Identification query.
        """
        try:
            logging.debug('{:s}: Get IDN'.format(__name__))
            return str(self.query('V'))
        except Exception as e:
            logging.error('{!s}: Cannot get IDN'.format(__name__))
            raise type(e)('{!s}: Cannot get IDN\n{!s}'.format(__name__, e))
    
    def get_status(self, *args):
        """
        Gets status entries.
        
        Meanings are:
            X: SYSTEM STATUS (Always zero currently)
            A: AUTO/MAN STATUS (n as for A COMMAND but see below)
            C: LOC/REM/LOCK STATUS (n as for C COMMAND)
            S: SWEEP STATUS (nn=0-32 as follows)
                nn=0 SWEEP NOT RUNNING
                nn=2P-1 SWEEPING to step P
                nn=2P HOLDING at step P
            H: CONTROL SENSOR (n as for H COMMAND)
            L: AUTO-PID STATUS (n as for L COMMAND)
        
        Parameters
        ----------
        *args: (str)
            variables of interest
        
        Returns
        -------
        status: (dict)
            All or selected entries of status query.
        """
        # Corresponding Command: XnAnCnSnnHnLn = X
        try:
            logging.debug('{:s}: Get status'.format(__name__))
            ans = self.query('X')
            status = {str(key): str(val) for key, val in zip(re.findall(r'[A-Za-z]+', ans), re.findall(r'\d+', ans))}
            if args:
                return {var: status[var] for var in args}
            else:
                return status
        except Exception as e:
            logging.error('{!s}: Cannot get status'.format(__name__))
            raise type(e)('{!s}: Cannot get status{!s}'.format(__name__, e))
        
    def set_control(self, remote=1, unlocked=1):
        """
        Sets the device into LOCAL or REMOTE and determines whether the LOC/REM button is LOCKED or active.
        
        Parameters
        ----------
        remote: (int)
            0 for local, 1 for remote
        unlocked: (int)
            0 to lock, 1 to unlock
        
        Returns
        -------
        None
        """
        # Corresponding Command: Cn
        try:
            logging.debug('{:s}: Set control to {:r}, {:r}'.format(__name__, remote, unlocked))
            self.write("C{:d}".format(int(str(remote) + str(unlocked), 2)))
        except Exception as e:
            logging.error('{!s}: Cannot set control to {!r}, {!r}'.format(__name__, remote, unlocked))
            raise type(e)('{!s}: Cannot set control o {!r}, {!r}\n{!s}'.format(__name__, remote, unlocked, e))
        return
    
    def set_display(self, channel):
        """
        Sets the front panel display to channel <channel>.
        
        Parameters
        ----------
        channel: (int)
            Channel to read. Possible channels are:
                0: SET TEMPERATURE
                1: SENSOR 1 TEMPERATURE
                2: SENSOR 2 TEMPERATURE
                3: SENSOR 3 TEMPERATURE
                4: TEMPERATURE ERROR
                5: HEATER O/P (as %)
                6: HEATER O/P (as V)
                7: GAS FLOW O/P (a.u.)
                8: PROPORTIONAL BAND
                9: INTEGRAL ACTION TIME
                10: DERIVATIVE ACTION TIME
        
        Returns
        -------
        None
        """
        # Corresponding Command: Fnn
        assert type(channel) == int, 'Argument must be an integer.'
        assert channel in range(0,11), 'Argument is not a valid number.'
        try:
            logging.debug('{:s}: Get value of channel {:d}'.format(__name__, channel))
            self.write('F{:d}'.format(channel))
        except Exception as e:
            logging.error('{!s}: Cannot set front pannel to channel {!s}'.format(__name__, channel))
            raise type(e)('{!s}: Cannot set front pannel to channel {!s}\n{!s}'.format(__name__, channel, e))
    
    def get_value(self, channel, display=False):
        """
        Gets the current value of channel <channel>.
        
        Parameters
        ----------
        channel: (int)
            Channel to read. Possible channels are:
                0: SET TEMPERATURE
                1: SENSOR 1 TEMPERATURE
                2: SENSOR 2 TEMPERATURE
                3: SENSOR 3 TEMPERATURE
                4: TEMPERATURE ERROR
                5: HEATER O/P (as %)
                6: HEATER O/P (as V)
                7: GAS FLOW O/P (a.u.)
                8: PROPORTIONAL BAND
                9: INTEGRAL ACTION TIME
                10: DERIVATIVE ACTION TIME
        
        Returns
        -------
        val: (float)
            answer of queried variable.
        """
        # Corresponding Command: Rn
        assert type(channel) == int, 'Argument must be an integer.'
        assert channel in range(0,11), 'Argument is not a valid number.'
        try:
            logging.debug('{:s}: Get value of channel {:d}'.format(__name__, channel))
            if display: self.set_display(channel=channel)
            return float(self.query('R{:d}'.format(channel)).strip('R+'))
        except Exception as e:
            logging.error('{!s}: Cannot get value of channel  {!s}'.format(__name__, channel))
            raise type(e)('{!s}: Cannot get value of channel  {!s}\n{!s}'.format(__name__, channel, e))
    
    def get_Temp(self, sensor=2):
        """
        Gets temperature of thermometer channel <channel>.
        
        Parameters
        ----------
        sensor: int
            Number of sensor of interest. Must be 1 (Sorb), 2 (RuOx) or 3 (Cernox). Default is 2.
        
        Returns
        -------
        temp: float
            Temperature in K.
        """
        # Corresponding Command: Rn
        assert type(sensor) == int, 'Argument must be an integer.'
        assert sensor in range(1, 4), 'Argument is not a valid number.'
        try:
            logging.debug('{:s}: Get temperature of sensor {:d}'.format(__name__, sensor))
            return float(self.query('R{:d}'.format(sensor)).strip('R+'))
        except Exception as e:
            logging.error('{!s}: Cannot get temperature of sensor {!s}'.format(__name__, sensor))
            raise type(e)('{!s}: Cannot get temperature of sensor {!s}\n{!s}'.format(__name__, sensor, e))
        
    def set_proportional(self, proportional=0):
        """
        Sets the proportional band.
        
        Parameters
        ----------
        proportional: (float)
            Proportional band, in steps of 0.0001K.
        
        Returns
        -------
        None
        """
        # Corresponding Command: Pnnnn
        try:
            logging.debug('{:s}: Set proportional band to {:f}'.format(__name__, proportional))
            self.write('P{}'.format(proportional))
        except Exception as e:
            logging.error('{!s}: Cannot set proportional band to {!s}'.format(__name__, proportional))
            raise type(e)('{!s}: Cannot set proportional band to {!s}\n{!s}'.format(__name__, proportional, e))
        return
        
    def set_integral(self, integral=0):
        """
        Sets the integral action time.
        
        Parameters
        ----------
        integral: (float)
            Integral action time, in steps of 0.1 minute. Ranges from 0 to 140 minutes.
        
        Returns
        -------
        None
        """
        # Corresponding Command: Innnn
        try:
            logging.debug('{:s}: Set integral action time to {:f}'.format(__name__, integral))
            self.write('I{}'.format(integral))
        except Exception as e:
            logging.error('{!s}: Cannot set integral action time to {!s}'.format(__name__, integral))
            raise type(e)('{!s}: Cannot set integral action time to {!s}\n{!s}'.format(__name__, integral, e))
        return
        
    def set_derivative(self, derivative=0):
        """
        Sets the derivative action time.
        
        Parameters
        ----------
        derivative: (float)
            Derivative action time. Ranges from 0 to 273 minutes.
        
        Returns
        -------
        None
        """
        # Corresponding Command: Dnnnn
        try:
            logging.debug('{:s}: Set derivative action time to {:f}'.format(__name__, derivative))
            self.write('D{}'.format(derivative))
        except Exception as e:
            logging.error('{!s}: Cannot set derivative action time to {!s}'.format(__name__, derivative))
            raise type(e)('{!s}: Cannot set derivative action time to {!s}\n{!s}'.format(__name__, derivative, e))
        return

    def set_temperature(self, temperature=0.010):
        """
        Change the temperature set point.
        
        Parameters
        ----------
        temperature: (float)
            Temperature set point in Kelvin. Default is 0.010
        
        Returns
        -------
        None
        """
        # Corresponding Command: Tnnnnn
        assert type(temperature) in [int, float], 'argument must be a number'
        try:
            logging.debug('{:s}: Set temperature set point to {:f}'.format(__name__, temperature))
            self.write('T{:d}'.format(int(1000*temperature)))
        except Exception as e:
            logging.error('{!s}: Cannot set temperature set point to {!s}'.format(__name__, temperature))
            raise type(e)('{!s}: Cannot set temperature set point to {!s}\n{!s}'.format(__name__, temperature, e))
        return
        
    def set_heater_sensor(self, sensor=1):
        """
        Selects the heater sensor.
        
        Parameters
        ----------
        sensor: (int)
            Heater sensor corresponding to the three input channels. Should be 1, 2, or 3, corresponding to the heater on the front panel.
        
        Returns
        -------
        None
        """
        # Corresponding Command: Hn
        assert sensor in [1,2,3], 'Heater not on list.'
        try:
            logging.debug('{:s}: Set heater sensor. to {:f}'.format(__name__, sensor))
            self.write('H{}'.format(sensor))
        except Exception as e:
            logging.error('{!s}: Cannot set heater sensor. to {!s}'.format(__name__, sensor))
            raise type(e)('{!s}: Cannot set heater sensor. to {!s}\n{!s}'.format(__name__, sensor, e))
        return
            
    def set_heater_maximum(self, val):
        """
        Sets the maximum heater voltage that ITC503 may deliver
        
        Parameters
        ----------
        val: float
            The maximum heater voltage is specified as a decimal number with a resolution of 0.1 volt, and is approximate. 0 may be used to specify a dynamically varying limit.
        
        Returns
        -------
        None
        """
        # Corresponding Command: Mnnn
        try:
            logging.debug('{:s}: Set maximum heater voltage to {:f}'.format(__name__, val))
            self.write('M{:04.1F}'.format(val))
        except Exception as e:
            logging.error('{!s}: Cannot set maximum heater voltage to {!s}'.format(__name__, val))
            raise type(e)('{!s}: Cannot set maximum heater voltage to {!s}\n{!s}'.format(__name__, val, e))
        return
    
    def set_heater_output(self, val):
        """
        Sets the heater output level <val>.
        
        Parameters
        ----------
        val: float
            Sets the percent of the maximum heater output in units of 0.1%. Ranges from 0 to 99.9.
        
        Returns
        -------
        None
        """
        # Corresponding Command: Onnn
        try:
            logging.debug('{:s}: Set heater output to {:f}'.format(__name__, val))
            self.query('O{:04.1F}'.format(val))
        except Exception as e:
            logging.error('{!s}: Cannot set heater output to {!s}'.format(__name__, val))
            raise type(e)('{!s}: Cannot set heater output to {!s}\n{!s}'.format(__name__, val, e))
        return
    
    def set_gas_output(self, val=0):
        """
        Sets the gas (needle valve) output level <gas_output>.
        
        Parameters
        ----------
        val: (float)
            Sets the percent of the maximum gas output in units of 0.1%. Ranges from 0 to 99.9.
        
        Returns
        -------
        None
        """
        # Corresponding Command: Gnnn
        try:
            logging.debug('{:s}: Set gas output to {:f}'.format(__name__, val))
            self.query('G{:04.1F}'.format(val))
        except Exception as e:
            logging.error('{!s}: Cannot set gas output to {!s}'.format(__name__, val))
            raise type(e)('{!s}: Cannot set gas output to {!s}\n{!s}'.format(__name__, val, e))
        return
    
    def set_auto_control(self, heater_auto=0, gas_auto=0):
        """
        Sets automatic control for heater and gas flow (needle valve).
        
        Parameters
        ----------
        heater_auto: (bool)
            Status of heater function.
        gas_auto: (bool)
            Status of gas flow control function.
        
        Returns
        -------
        None
        """
        # Corresponding Command: An
        try:
            logging.debug('{:s}: Set automatic control for heater to {:d} and gas flow to {:d}'.format(__name__, heater_auto, gas_auto))
            self.write('A{}'.format(int(str(int(gas_auto))+str(int(heater_auto)), 2)))
        except Exception as e:
            logging.error('{!s}: Cannot set automatic control for heater to {!s} and gas flow to {!s}'.format(__name__, heater_auto, gas_auto))
            raise type(e)('{!s}: Cannot set automatic control for heater to {!s} and gas flow to {!s}\n{!s}'.format(__name__, heater_auto, gas_auto, e))
        return
    
    def set_sweeps(self, sweep_parameters):
        """
        Sets the parameters for all sweeps.

        This fills up a dictionary with all the possible steps in a sweep. If a step number is not found in the sweep_parametersdictionary, then it will create the sweep step with all parameters set to 0.
        
        Parameters
        ----------
        sweep_parameters: (dict)
            Keys are the step numbers and range fron 1 to 16. The value of each key is a dictionary whose keys are the parameters in the sweep table (see _setSweepStep).
        
        Returns
        -------
        None
        """
        for step in range(1,17):
            if step in sweep_parameters.keys():
                self._set_sweep_step(step, sweep_parameters[step])
            else:
                self._set_sweep_step(step, {'set_point': 0, 'sweep_time': 0, 'hold_time': 0 })  # default is 0
    
    def _set_sweep_step(self, sweep_step, sweep_table):
        """
        Sets the parameters for a sweep step by setting the step pointer (x) to the proper step, the step parameters (y1, y2, y3) to the values dictated by the sweep_table and resetting the x and y pointers to 0.
        
        Parameters
        ----------
        sweep_step: (int)
            The sweep step to be modified. Ranges from 1 to 16
        sweep_table: (int)
            A dictionary of parameters describing the sweep. Keys are set_point, sweep_time and hold_time.
        
        Returns
        -------
        None
        """
        self.write('x{}'.format(sweep_step))

        self.write('y1')
        self.write('s{}'.format(sweep_table['set_point']))

        self.write('y2')
        self.write('s{}'.format(sweep_table['sweep_time']))

        self.write('y3')
        self.write('s{}'.format(sweep_table['hold_time']))

        self._reset_sweep_table_pointers()

    def _reset_sweep_table_pointers(self):
        """
        Resets the table pointers to x=0 and y=0 to prevent accidental sweep table changes.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        """
        self.write('x0')
        self.write('y0')
    


# Agilent_VNA_E5071C driver, P. Macha, modified by M. Weides July 2013, J. Braumueller 2015
# Adapted to Keysight VNA by A. Schneider and L. Gurenhaupt 2016
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from qkit.core.instrument_base import Instrument
from qkit import visa
import types
import logging
from time import sleep
import numpy

class Keysight_VNA_P9375A(Instrument):
    '''
    This is the python driver for the Anritsu MS4642A Vector Network Analyzer

    Usage:
    Initialise with
    <name> = instruments.create('<name>', address='<GPIB address>', reset=<bool>)
    
    '''

    def __init__(self, name, address, channel_index = 1):
        '''
        Initializes 

        Input:
            name (string)    : name of the instrument
            address (string) : GPIB address
        '''
        
        logging.info(__name__ + ' : Initializing instrument')
        Instrument.__init__(self, name, tags=['physical'])

        self._address = address
        self._visainstrument = visa.instrument(self._address)
        self._visainstrument.timeout = 10000
        self._zerospan = False
        self._freqpoints = 0
        self._ci = channel_index 
        self._active_trace = 1
        self._pi = 2 # port_index, similar to self._ci
        self._start = 0
        self._stop = 0
        self._nop = 0
        
        # Turn on "fixturing" such that data fetched is corrected for delay
        self._visainstrument.write("CALC:FSIM:STAT ON")

        # Implement parameters
        self.add_parameter('nop', type=int,
            flags=Instrument.FLAG_GETSET,
            minval=2, maxval=100001,
            tags=['sweep'])
            
        self.add_parameter('bandwidth', type=float,
            flags=Instrument.FLAG_GETSET,
            minval=0, maxval=1e9,
            units='Hz', tags=['sweep']) 

        self.add_parameter('averages', type=int,
            flags=Instrument.FLAG_GETSET,
            minval=1, maxval=1024, tags=['sweep'])                    

        self.add_parameter('Average', type=bool,
            flags=Instrument.FLAG_GETSET)   
                    
        self.add_parameter('centerfreq', type=float,
            flags=Instrument.FLAG_GETSET,
            minval=0, maxval=20e9,
            units='Hz', tags=['sweep'])

        self.add_parameter('cwfreq', type=float,
            flags=Instrument.FLAG_GETSET,
            minval=0, maxval=20e9,
            units='Hz', tags=['sweep'])
            
        self.add_parameter('startfreq', type=float,
            flags=Instrument.FLAG_GETSET,
            minval=0, maxval=20e9,
            units='Hz', tags=['sweep'])            
            
        self.add_parameter('stopfreq', type=float,
            flags=Instrument.FLAG_GETSET,
            minval=0, maxval=20e9,
            units='Hz', tags=['sweep'])                        
            
        self.add_parameter('span', type=float,
            flags=Instrument.FLAG_GETSET,
            minval=0, maxval=20e9,
            units='Hz', tags=['sweep'])        
            
        self.add_parameter('power', type=float,
            flags=Instrument.FLAG_GETSET,
            minval=-80, maxval=20,
            units='dBm', tags=['sweep'])

        self.add_parameter('startpower', type=float,
            flags=Instrument.FLAG_GETSET,
            minval=-100, maxval=20,
            units='dBm')

        self.add_parameter('stoppower', type=float,
            flags=Instrument.FLAG_GETSET,
            minval=-100, maxval=20,
            units='dBm')

        self.add_parameter('cw', type=bool,
            flags=Instrument.FLAG_GETSET)

        self.add_parameter('zerospan', type=bool,
            flags=Instrument.FLAG_GETSET)
            
        self.add_parameter('channel_index', type=int,
            flags=Instrument.FLAG_GETSET)

        self.add_parameter('sweeptime', type=float,   #added by MW
            flags=Instrument.FLAG_GET,
            minval=0, maxval=1e3,
            units='s', tags=['sweep'])
            
        self.add_parameter('sweeptime_averages', type=float,   #JB
            flags=Instrument.FLAG_GET,
            minval=0, maxval=1e3,
            units='s', tags=['sweep'])
    
        self.add_parameter('edel', type=float, # legacy name for parameter. This corresponds to the VNA's port extension values.
            flags=Instrument.FLAG_GETSET, 
            minval=-10, maxval=10,
            units='s', tags=['sweep'],
            channels=(1, self._pi), channel_prefix = 'port%d_') # the channel option for qtlab's Instument class is used here to easily address the two VNA ports
  
        self.add_parameter('edel_status', type=bool, # legacy name for parameter. This corresponds to the VNA's port extension values.
            flags=Instrument.FLAG_GETSET)
                   
        self.add_parameter('sweep_mode', type=str,  #JDB This parameter switches on/off hold. The hold function below does the same job, this is just for code compatibility to the agilent and anritsu drivers.
            flags=Instrument.FLAG_GETSET,tags=['sweep']) 
                    
        self.add_parameter('sweep_type', type=str,
            flags=Instrument.FLAG_GETSET,tags=['sweep'])

        self.add_parameter('active_trace', type=int,
            flags=Instrument.FLAG_GETSET)
        
        self.add_parameter('output', type=bool,
            flags=Instrument.FLAG_GETSET)   
        
        #Triggering Stuff
        self.add_parameter('trigger_source', type=str,
            flags=Instrument.FLAG_GETSET)
        
        # Implement functions
        self.add_function('get_freqpoints')
        self.add_function('get_tracedata')
        self.add_function('init')
        self.add_function('avg_clear')
        self.add_function('avg_status')
        self.add_function('def_trig')
        self.add_function('get_hold')
        self.add_function('hold')
        self.add_function('get_sweeptime')
        self.add_function('get_sweeptime_averages')
        self.add_function('pre_measurement')
        self.add_function('start_measurement')
        self.add_function('ready')
        self.add_function('post_measurement')

        self.get_all()
    
    def get_all(self):        
        self.get_nop()
        self.get_power()
        self.get_centerfreq()
        self.get_startfreq()
        self.get_stopfreq()
        self.get_span()
        self.get_bandwidth()
        self.get_trigger_source()
        self.get_Average()
        self.get_averages()
        self.get_freqpoints()   
        self.get_channel_index()
        self.get_zerospan()
        self.get_sweeptime()
        self.get_sweeptime_averages()
        self.get_edel_status()
        self.get_cw()
        self.get_cwfreq()
        self.get_output()
        for port in range(self._pi):
            self.get('port%d_edel' % (port+1))
        
    ###
    #Communication with device
    ###
    
    
    def hold(self, status):     # added MW July 13
        self._visainstrument.write(":TRIG:SOUR IMM")
        if status:
            self._visainstrument.write('INIT:CONT OFF')
            self._visainstrument.write('SENS:SWE:MODE HOLD') # set sweep mode to hold
            self._visainstrument.write('OUTP OFF') # set RF output off
        else:
            self._visainstrument.write('INIT:CONT ON')
            self._visainstrument.write('SENS:SWE:MODE CONT') # set sweep mode to hold
            self._visainstrument.write('OUTP ON') # set RF output on

    def get_hold(self):     # added MW July 13
        ###hold==False only if rf power is on,  sweep mode is continuous and trigger source is immediate (interal continuous triggering). 
        if int(self._visainstrument.query('INIT%i:CONT?'%(self._ci))) and self._visainstrument.query(':SENS%i:SWE:MODE?'%(self._ci)) == 'CONT\n' and int(self._visainstrument.query('OUTP?')):
            self._hold=False
        elif not int(self._visainstrument.query('INIT%i:CONT?'%(self._ci))) and self._visainstrument.query(':SENS%i:SWE:MODE?'%(self._ci)) == 'HOLD\n' and not int(self._visainstrument.query('OUTP?')):
            self._hold=True
        else:
            self._hold=None
        return self._hold 
    
    def init(self):
        if self._zerospan:
          self._visainstrument.write('INIT1;*wai')
        else:
          if self.get_Average():
            for i in range(self.get_averages()):            
              self._visainstrument.write('INIT1;*wai')
          else:
              self._visainstrument.write('INIT1;*wai')   

    def def_trig(self):
        self._visainstrument.write(':TRIG:AVER ON')
        self._visainstrument.write(':TRIG:SOUR bus')
        
    def avg_clear(self):
        #self._visainstrument.write(':TRIG:SING')
        #self._visainstrument.write(':SENSe%i:AVERage:CLEar'%(self._ci))
        self._visainstrument.write(':SENS%i:AVER:CLE' %(self._ci))

    def avg_status(self):
        return 0 == (int(self._visainstrument.query('STAT:OPER:COND?')) & (1<<4))
        #return int(self._visainstrument.query(':SENS%i:AVER:COUNT?' %(self._ci)))    
        
    def get_tracedata(self, format = 'AmpPha', single=False, averages=1.):
        
        
        '''
        Get the data of the current trace

        Input:
            format (string) : 'AmpPha': Amp in dB and Phase, 'RealImag',

        Output:
            'AmpPha':_ Amplitude and Phase
        '''
        
        
        if single==True:        #added MW July. 
            #print('single shot readout')
            self._visainstrument.write('TRIG:SOUR IMM') #added MW July 2013. start single sweep.
            self._visainstrument.write('INIT%i:CONT ON'%(self._ci)) #added MW July 2013. start single sweep.
            self.hold(True)
            sleep(float(self._visainstrument.query('SENS1:SWE:TIME?'))) 
        
        #sleep(0.1) # required to avoid timing issues    MW August 2013   ???
            
        #self._visainstrument.write(':FORMAT REALform; FORMat:BORDer SWAP;')
        #data = self._visainstrument.query_binary_values( "CALCulate:DATA? SDATA",format = visa.single)
        #data = self._visainstrument.query_binary_values(':FORMAT REAL,32;*CLS;CALC1:DATA:NSW? SDAT,1;*OPC',format=1)      
        #data = self._visainstrument.query_binary_values('FORM:DATA REAL; FORM:BORD SWAPPED; CALC%i:SEL:DATA:SDAT?'%(self._ci), format = visa.double)  
        self._visainstrument.write('FORM:DATA REAL')
        self._visainstrument.write('FORM:BORD SWAPPED') #SWAPPED
        #data = self._visainstrument.query_binary_values('CALC%d:SEL:DATA:SDAT?'%(self._ci), format = visa.double)              
        data = self._visainstrument.query_binary_values('CALC%i:MEAS:DATA:SDAT?'%(self._ci))
        data_size = numpy.size(data)
        datareal = numpy.array(data[0:data_size:2])
        dataimag = numpy.array(data[1:data_size:2])
          
        if format == 'RealImag':
          if self._zerospan:
            return numpy.mean(datareal), numpy.mean(dataimag)
          else:
            return datareal, dataimag
        elif format == 'AmpPha':
          if self._zerospan:
            datareal = numpy.mean(datareal)
            dataimag = numpy.mean(dataimag)
            dataamp = numpy.sqrt(datareal*datareal+dataimag*dataimag)
            for i in numpy.arange(len(datareal)):    #added MW July 2013
                if datareal[i]>=0 and dataimag[i] >=0:   #1. quadrant
                    datapha = numpy.arctan(dataimag[i]/datareal[i])
                elif  datareal[i] <0 and dataimag[i] >=0:  #2. quadrant
                    datapha = numpy.arctan(dataimag[i]/datareal[i])+ numpy.pi
                elif  datareal[i] <0 and dataimag[i] <0:  #3. quadrant
                    datapha = numpy.arctan(dataimag[i]/datareal[i])- numpy.pi
                elif  datareal[i] >=0 and dataimag[i]<0:   #4. quadrant
                    datapha = numpy.arctan(dataimag[i]/datareal[i])
                    
            return dataamp, datapha
          else:
            dataamp = numpy.sqrt(datareal*datareal+dataimag*dataimag)
            datapha = numpy.arctan2(dataimag,datareal)
            if self.get_cw():
                # in cw mode the vna performs a power sweep with 2 power values. the driver sets them to the same level,
                # but the returned data may cause a dimension problem.
                # therefore we only take one datapoint or average them.
                # Spectroscopy requires a sized object, so there must be [] around the values
                dataamp, datapha = numpy.array([numpy.mean(dataamp)]), numpy.array([numpy.mean(datapha)])
            return dataamp, datapha
        else:
          raise ValueError('get_tracedata(): Format must be AmpPha or RealImag') 
      
    def get_freqpoints(self, query = False):      
      if query == True:        
        self._freqpoints = self._visainstrument.query_binary_values('FORM:DATA REAL; FORM:BORD SWAPPED; :SENS%i:FREQ:DATA?'%(self._ci), format = visa.double)
      
        #self._freqpoints = numpy.array(self._visainstrument.query_binary_values('SENS%i:FREQ:DATA:SDAT?'%self._ci,format=1)) / 1e9
        #self._freqpoints = numpy.array(self._visainstrument.query_binary_values(':FORMAT REAL,32;*CLS;CALC1:DATA:STIM?;*OPC',format=1)) / 1e9
      elif self.get_cw():
          self._freqpoints = numpy.atleast_1d(self.get_cwfreq())
      else:
          self._freqpoints = numpy.linspace(self._start,self._stop,self._nop)
      return self._freqpoints

    ###
    # SET and GET functions
    ###
    
    def do_set_sweep_mode(self, mode):
        '''
        select the sweep mode from 'hold', 'cont', single'
        single means only one single trace, not all the averages even if averages
         larger than 1 and Average==True

        Input:
            source (string) : HOLD | CONTinuous | SINGle 

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting sweep mode "%s"' % mode)
        if mode.upper() in ['HOLD', 'CONT', 'SING']:
            self._visainstrument.write('SENS:SWE:MODE %s' % mode.upper())        
        else:
            raise ValueError('set_trigger_source(): must be HOLD | CONTinuous | SINGle ')
        
           
    def do_get_sweep_mode(self):
        return self._visainstrument.query(':SENS%i:SWE:MODE?'%(self._ci))
    
    def do_set_output(self, status):
        '''
        sets the RF power on or off

        Input:
            status (boolean)

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting RF output to "%s"' % (status))
        if status:
            self._visainstrument.write('OUTP ON')
        elif status == False:
            self._visainstrument.write('OUTP OFF')
        else:
            raise ValueError('set_output(): can only set True or False')               
    def do_get_output(self):
        '''
        Get status of RF output

        Input:
            None

        Output:
            Status of RF output (boolean)
        '''
        logging.debug(__name__ + ' : getting average status')
        return bool(int(self._visainstrument.query('OUTP?')))
    
    def do_set_nop(self, nop):
        '''
        Set Number of Points (nop) for sweep

        Input:
            nop (int) : Number of Points

        Output:
            None
        '''

        logging.debug(__name__ + ' : setting Number of Points to %s ' % (nop))
        if self._zerospan:
          print('in zerospan mode, nop is 1')
        else:
          cw = self.get_cw()
          if(nop == 1) and (not cw):
              logging.info(__name__ + 'nop == 1 is only supported in CW mode.')
              self.set_cw(True)
          else:
              self._visainstrument.write(':SENS%i:SWE:POIN %i' %(self._ci,nop))
          """
          Comment below is copy from Anritsu driver
          if(cw):
            self._visainstrument.write(':SENS%i:SWE:CW:POIN %i' %(self._ci,nop))
          else:
            self._visainstrument.write(':SENS%i:SWE:POIN %i' %(self._ci,nop))
          """
          self._nop = nop
          self.get_freqpoints() #Update List of frequency points  
        
    def do_get_nop(self):
        '''
        Get Number of Points (nop) for sweep

        Input:
            None
        Output:
            nop (int)
        '''
        logging.debug(__name__ + ' : getting Number of Points')
        if self._zerospan or self.get_cw():
          self._nop = 1
        else:
          self._nop = int(self._visainstrument.query(':SENS%i:SWE:POIN?' %(self._ci)))    
        return self._nop 
    
    def do_set_Average(self, status):
        '''
        Set status of Average

        Input:
            status (boolean)

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting Average to "%s"' % (status))
        if status:
            status = 'ON'
            self._visainstrument.write('SENS%i:AVER:STAT %s' % (self._ci,status))
        elif status == False:
            status = 'OFF'
            self._visainstrument.write('SENS%i:AVER:STAT %s' % (self._ci,status))
        else:
            raise ValueError('set_Average(): can only set True or False')               
    def do_get_Average(self):
        '''
        Get status of Average

        Input:
            None

        Output:
            Status of Averaging (boolean)
        '''
        logging.debug(__name__ + ' : getting average status')
        return bool(int(self._visainstrument.query('SENS%i:AVER:STAT?' %(self._ci))))
                    
    def do_set_averages(self, av):
        '''
        Set number of averages

        Input:
            av (int) : Number of averages

        Output:
            None
        '''
        if self._zerospan == False:
            logging.debug(__name__ + ' : setting Number of averages to %i ' % (av))
            self._visainstrument.write('SENS%i:AVER:COUN %i' % (self._ci,av))
        else:
            self._visainstrument.write('SWE%i:POIN %.1f' % (self._ci,av))
            
    def do_get_averages(self):
        '''
        Get number of averages

        Input:
            None
        Output:
            number of averages
        '''
        logging.debug(__name__ + ' : getting Number of Averages')
        if self._zerospan:
          return int(self._visainstrument.query('SWE%i:POIN?' % self._ci))
        else:
          return int(self._visainstrument.query('SENS%i:AVER:COUN?' % self._ci))
                
    def do_set_power(self,pow):
        '''
        Set probe power

        Input:
            pow (float) : Power in dBm

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting power to %s dBm' % pow)
        if self.get_cw():
            self.set_startpower(pow)
            self.set_stoppower(pow)
        else:
            self._visainstrument.write('SOUR%i:POW1:LEV:IMM:AMPL %.1f' % (self._ci,pow))    
    def do_get_power(self):
        '''
        Get probe power

        Input:
            None

        Output:
            pow (float) : Power in dBm
        '''
        logging.debug(__name__ + ' : getting power')
        if self.get_cw():
            return self.get_startpower()
        else:
            return float(self._visainstrument.query('SOUR%i:POW1:LEV:IMM:AMPL?' % (self._ci)))

    def do_set_startpower(self,pow):
        '''
        Set probe power

        Input:
            pow (float) : Power in dBm

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting startpower to %s dBm' % pow)
        self._visainstrument.write('SOUR%i:POW:START %.1f' % (self._ci,pow))    
    def do_get_startpower(self):
        '''
        Get probe power

        Input:
            None

        Output:
            pow (float) : Power in dBm
        '''
        logging.debug(__name__ + ' : getting startpower')
        return float(self._visainstrument.query('SOUR%i:POW:START?' % (self._ci)))

    def do_set_stoppower(self,pow):
        '''
        Set probe power

        Input:
            pow (float) : Power in dBm

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting stoppower to %s dBm' % pow)
        self._visainstrument.write('SOUR%i:POW:STOP %.1f' % (self._ci,pow))    
    def do_get_stoppower(self):
        '''
        Get probe power

        Input:
            None

        Output:
            pow (float) : Power in dBm
        '''
        logging.debug(__name__ + ' : getting stoppower')
        return float(self._visainstrument.query('SOUR%i:POW:STOP?' % (self._ci)))
                
    def do_set_centerfreq(self,cf):
        '''
        Set the center frequency

        Input:
            cf (float) :Center Frequency in Hz

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting center frequency to %s' % cf)
        self._visainstrument.write('SENS%i:FREQ:CENT %f' % (self._ci,cf))
        self.get_startfreq()
        self.get_stopfreq()
        self.get_span()
    def do_get_centerfreq(self):
        '''
        Get the center frequency

        Input:
            None

        Output:
            cf (float) :Center Frequency in Hz
        '''
        logging.debug(__name__ + ' : getting center frequency')
        return  float(self._visainstrument.query('SENS%i:FREQ:CENT?'%(self._ci)))

    def do_set_cwfreq(self, cf):
        ''' set cw frequency '''
        self._visainstrument.write('SENS%i:FREQ:CW %f' % (self._ci,cf))
    
    def do_get_cwfreq(self):
        ''' get cw frequency '''
        return float(self._visainstrument.query('SENS%i:FREQ:CW?'%(self._ci)))
        
    def do_set_span(self,span):
        '''
        Set Span

        Input:
            span (float) : Span in KHz

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting span to %s Hz' % span)
        self._visainstrument.write('SENS%i:FREQ:SPAN %i' % (self._ci,span))   
        self.get_startfreq()
        self.get_stopfreq()
        self.get_centerfreq()
        
    def do_get_span(self):
        '''
        Get Span
        
        Input:
            None

        Output:
            span (float) : Span in Hz
        '''
        #logging.debug(__name__ + ' : getting center frequency')
        span = self._visainstrument.query('SENS%i:FREQ:SPAN?' % (self._ci) ) #float( self.ask('SENS1:FREQ:SPAN?'))
        return span

    def do_get_sweeptime_averages(self):###JB
        '''
        Get sweeptime
        
        Input:
            None

        Output:
            sweep time (float) times number of averages: sec
        '''
        return self.get_sweeptime() * self.get_averages()
        
    def do_get_sweeptime(self):  #added MW July 2013
        '''
        Get sweeptime
        
        Input:
            None

        Output:
            sweep time (float) : sec
        '''
        logging.debug(__name__ + ' : getting sweep time')
        self._sweep=float(self._visainstrument.query('SENS1:SWE:TIME?'))
        return self._sweep


    def do_set_edel(self, val,channel):  # MP 04/2017

        '''
        Set electrical delay

        '''
        logging.debug(__name__ + ' : setting port %s extension to %s sec' % (channel, val))
        self._visainstrument.write('SENS1:CORR:EXT:PORT%i:TIME %.12f' % (channel, val))
            
    
    def do_get_edel(self, channel):   # MP 04/2017

        '''
        Get electrical delay

        '''
        logging.debug(__name__ + ' : getting port %s extension' % channel)
        self._edel = float(self._visainstrument.query('SENS1:CORR:EXT:PORT%i:TIME?'% channel))
        return  self._edel   
        
    def do_set_edel_status(self, status):   # MP 04/2017

        '''
        Set electrical delay

        '''
        logging.debug(__name__ + ' : setting port extension status to %s' % (status))
        self._visainstrument.write('SENS.CORR.EXT.STAT %i' % (status))
            
    
    def do_get_edel_status(self):   # MP 04/2017

        '''
        Get electrical delay

        '''
        logging.debug(__name__ + ' :  port extension status')
        return  self._visainstrument.query('SENS:CORR:EXT:STAT?')
        
        
    def do_set_startfreq(self,val):
        '''
        Set Start frequency

        Input:
            span (float) : Frequency in Hz

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting start freq to %s Hz' % val)
        self._visainstrument.write('SENS%i:FREQ:STAR %f' % (self._ci,val))   
        self._start = val
        self.get_centerfreq()
        self.get_stopfreq()
        self.get_span()
        
    def do_get_startfreq(self):
        '''
        Get Start frequency
        
        Input:
            None

        Output:
            span (float) : Start Frequency in Hz
        '''
        logging.debug(__name__ + ' : getting start frequency')
        self._start = float(self._visainstrument.query('SENS%i:FREQ:STAR?' % (self._ci)))
        return  self._start

    def do_set_stopfreq(self,val):
        '''
        Set STop frequency

        Input:
            val (float) : Stop Frequency in Hz

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting stop freq to %s Hz' % val)
        self._visainstrument.write('SENS%i:FREQ:STOP %f' % (self._ci,val))  
        self._stop = val
        self.get_startfreq()
        self.get_centerfreq()
        self.get_span()
    def do_get_stopfreq(self):
        '''
        Get Stop frequency
        
        Input:
            None

        Output:
            val (float) : Start Frequency in Hz
        '''
        logging.debug(__name__ + ' : getting stop frequency')
        self._stop = float(self._visainstrument.query('SENS%i:FREQ:STOP?' %(self._ci) ))
        return  self._stop

    def do_set_bandwidth(self,band):
        '''
        Set Bandwidth

        Input:
            band (float) : Bandwidth in Hz

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting bandwidth to %s Hz' % (band))
        self._visainstrument.write('SENS%i:BWID:RES %i' % (self._ci,band))
    def do_get_bandwidth(self):
        '''
        Get Bandwidth

        Input:
            None

        Output:
            band (float) : Bandwidth in Hz
        '''
        logging.debug(__name__ + ' : getting bandwidth')
        # getting value from instrument
        return  float(self._visainstrument.query('SENS%i:BWID:RES?'%self._ci))                
    
    def do_set_cw(self, val):
        '''
        Set instrument to CW (single frequency) mode and back
        '''
        if val:
            self.set_cwfreq(self.get_centerfreq())
            self.set_nop(2)
            self.set_startpower(self.get_power())
            self.set_stoppower(self.get_power())
            return self._visainstrument.write(':SENS%i:SWE:TYPE POW' %(self._ci))
        else:
            return self._visainstrument.write(':SENS%i:SWE:TYPE LIN' %(self._ci))
        
    def do_get_cw(self):
        '''
        retrieve CW mode status from device
        '''
        sweep_type = str(self._visainstrument.query(':SENS%i:SWE:TYPE?'%(self._ci))).rstrip()
        if sweep_type == 'POW': ret = True
        else: ret = False
        return ret

    def do_set_zerospan(self,val):
        '''
        Zerospan is a virtual "zerospan" mode. In Zerospan physical span is set to
        the minimal possible value (2Hz) and "averages" number of points is set.

        Input:
            val (bool) : True or False

        Output:
            None
        '''
        #logging.debug(__name__ + ' : setting status to "%s"' % status)
        if val not in [True, False]:
            raise ValueError('set_zerospan(): can only set True or False')        
        if val:
          self._oldnop = self.get_nop()
          self._oldspan = self.get_span()
          if self.get_span() > 0.002:
            Warning('Setting ZVL span to 2Hz for zerospan mode')            
            self.set_span(0.002)
            
        av = self.get_averages()
        self._zerospan = val
        if val:
            self.set_Average(False)
            self.set_averages(av)
            if av<2:
              av = 2
        else: 
          self.set_Average(True)
          self.set_span(self._oldspan)
          self.set_nop(self._oldnop)
          self.get_averages()
        self.get_nop()
               
    def do_get_zerospan(self):
        '''
        Check weather the virtual zerospan mode is turned on

        Input:
            None

        Output:
            val (bool) : True or False
        '''
        return self._zerospan


    def do_set_trigger_source(self,source):
        '''
        Set Trigger Mode

        Input:
            source (string) : IMMediate | MANual | EXTernal

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting trigger source to "%s"' % source)
        if source.upper() in ['IMM', 'MAN', 'EXT']:
            self._visainstrument.write('TRIG:SEQ:SOUR %s' % source.upper())        
        else:
            raise ValueError('set_trigger_source(): must be IMMediate | MANual | EXTernal')

    def do_get_trigger_source(self):
        '''
        Get Trigger Mode

        Input:
            None

        Output:
            source (string) : IMMediate  | MANual | EXTernal 
        '''
        logging.debug(__name__ + ' : getting trigger source')
        return str(self._visainstrument.query('TRIG:SEQ:SOUR?')).rstrip()
        

    def do_set_channel_index(self,val):
        '''
        Set the index of the channel to address.

        Input:
            val (int) : 1 .. number of active channels (max 16)

        Output:
            None
        '''
        logging.debug(__name__ + ' : setting channel index to "%i"' % int)
        nop = self._visainstrument.read('DISP:COUN?')
        if val < nop:
            self._ci = val 
        else:
            raise ValueError('set_channel_index(): index must be < nop channels')
    def do_get_channel_index(self):
        '''
        Get active channel

        Input:
            None

        Output:
            channel_index (int) : 1-16
        '''
        logging.debug(__name__ + ' : getting channel index')
        return self._ci
    
    def do_set_active_trace(self, trace):
        """
        Sets the active trace, which can then be readout
        :param trace: Number of the active trace
        :return: None
        """
        # TODO: catch error
        self._active_trace = trace
        self._visainstrument.write('CALC{}:PAR{}:SEL'.format(self._ci, trace))

    def do_get_active_trace(self):
        """
        :return: the active trace on the VNA
        """
        # TODO: ask device
        return self._active_trace

    def do_get_sweep_type(self):
        '''
        Get the Sweep Type

        Input:
            None

        Output:
            Sweep Type (string). One of
            LIN:    Frequency-based linear sweep
            LOG:    Frequency-based logarithmic sweep
            SEGM:   Segment-based sweep with frequency-based segments
            POW:    Power-based sweep with either Power-based sweep with a 
                CW frequency, or Power-based sweep with swept-frequency
        '''
        logging.debug(__name__ + ' : getting sweep type')
        
        return str(self._visainstrument.query('SENS%i:SWE:TYPE?' %(self._ci))).rstrip()
    
    def do_set_sweep_type(self,swtype):
        '''
        Set the Sweep Type
        Input:
            swtype (string):    One of
                LIN:    Frequency-based linear sweep
                LOG:    Frequency-based logarithmic sweep
                SEGM:   Segment-based sweep with frequency-based segments
                POW:    Power-based sweep with either Power-based sweep with a 
                    CW frequency, or Power-based sweep with swept-frequency

        Output:
            None            
        '''
        if swtype in ('LIN','LOG','SEGM','POW'):
            
            logging.debug(__name__ + ' : Setting sweep type to %s'%(swtype))
            return self._visainstrument.write('SENS%i:SWE:TYPE %s' %(self._ci,swtype))
            
        else:
            logging.debug(__name__ + ' : Illegal argument %s'%(swtype))
          
    def write(self,msg):
      return self._visainstrument.write(msg)    
    def ask(self,msg):
      return self._visainstrument.query(msg)
     
    def pre_measurement(self):
        '''
        Set everything needed for the measurement
        '''
        # self._visainstrument.write(":TRIG:SOUR MAN") #Only wait for software triggers
        # self._visainstrument.write('INIT:CONT ON')
        self.do_set_Average(True)
        pass
        
        
    def post_measurement(self):
        '''
        After a measurement, the VNA is in hold mode, and it can be difficult to start a measurement again from front panel.
        This function brings the VNA back to normal measuring operation.
        '''
        # self._visainstrument.write('INIT:CONT ON')
        # self.do_set_Average(False)
        self.hold(False)
        pass
        
      
    def start_measurement(self):
        '''
        This function is called at the beginning of each single measurement in the spectroscopy script.
        Here, it resets the averaging
        '''
        self._visainstrument.write('SENS1:AVER:STAT OFF')
        self._visainstrument.write(':SENS1:AVER:CLE')
        self._visainstrument.write('SENS1:AVER:STAT ON')
        # self._visainstrument.write('INIT:IMM') #go
        # self._visainstrument.write('*TRG') #go

    
    def ready(self):
        '''
        This is a proxy function, returning True when the VNA has finished the required number of averages.
        '''
        return (int(self._visainstrument.query(':STAT:OPER:AVER1:COND?')) & 2) == 2
        # return (int(self._visainstrument.query('STAT:OPER:DEV?')) & 16) == 16



from qkit.core.instrument_basev2 import ModernInstrument, QkitFunction
from qkit.drivers.AbstractVNA import AbstractVNA
from qkit.drivers.ZHInst_SHFSG import ZHInst_SHFSG
from qkit.drivers.ZHInst_UHFQA import ZHInst_UHFQA

import math

from zhinst.toolkit import TriggerMode, Alignment, SequenceType

import numpy as np

MAXIMUM_FREQUENCY = (8.5 + 0.65)*1e9
MAXIMUM_BANDWIDTH = (0.65)*1e9

MINIMUM_UHFQA_FREQUENCY = 50e6

class ZHInst_Virtual_VNAv2(ModernInstrument, AbstractVNA):

    """
    A Virtual VNA based on ZurichInstruments devices.
    Uses the SHFSG to generate the RF signal modulated by the IQ mixer connected to the UHFQA.

    We expect port 7 of the SHFSG to provide the RF signal.
    """
    def __init__(self, name, uhfqa: ZHInst_UHFQA, shfsg: ZHInst_SHFSG, **kwargs):
        super().__init__(name, tags=['virtual'], **kwargs)
        self.discover_capabilities()

        self.uhfqa = uhfqa
        self.shfsg = shfsg
        self.period = 10e-6
        self.repetitions = 512
        self.integration_length = 4096
        self.output_power = 0
        self.amplitude_factor = 1.0

        self._frequencies = np.ndarray([], dtype=float)
        self._i_q_data = np.ndarray([], dtype=float)

    @QkitFunction
    def configure_frequency_range(self, lower_freq: float, upper_freq: float, steps: int):
        assert 0 < lower_freq < upper_freq < MAXIMUM_FREQUENCY, "Frequency bounds invalid! (Too large? Order?)"
        assert 0 < steps, "Step count must be larger than zero!"
        self._frequencies = np.linspace(lower_freq, upper_freq, steps)

    @QkitFunction
    def configure_pulses(self, period = 10e-6, repetitions = 512, integration_length=4096):
        self.period = 10e-6
        self.repetitions = 512
        self.integration_length = 4096

    @QkitFunction
    def set_output_power(self, power):
        base_power = math.ceil(power / 5.0) * 5
        self.amplitude_factor = 10 ** (-0.5 * (base_power - power))
        self.output_power = base_power
        self.shfsg.nodetree.sgchannels[7].output.range(self.output_power)
        
        expected_voltage = 10**(0.5 * ((power / 10) - 1))
        self.uhfqa.nodetree.sigins[0].range(expected_voltage)
        self.uhfqa.nodetree.sigins[1].range(expected_voltage)

    @QkitFunction
    def get_freqpoints(self) -> np.ndarray:
        return self._frequencies

    @QkitFunction
    def get_tracedata(self, RealImag = None) -> tuple((np.ndarray, np.ndarray)):
        if not RealImag:
            signal = (self._i_q_data[:, 0] + 1.0j * self._i_q_data[:, 1])
            return np.abs(signal), np.angle(signal)
        else:
            return self._i_q_data[:, 0], self._i_q_data[:, 1]

    @QkitFunction
    def _get_shfsg_frequencies(self) -> tuple((float, float)):
        """
        Returns (SHF-Upconversion center frequency, Digital Modulation Frequency)
        """
        shfsg_out_freq = self.get_freqpoints()[0] - MINIMUM_UHFQA_FREQUENCY
        base_mod = int(shfsg_out_freq / 0.5e9) * 0.5e9
        return base_mod, shfsg_out_freq - base_mod

    @QkitFunction
    def pre_measurement(self):
        self.uhfqa.nodetree.qa.result.source(2)
        self.uhfqa.nodetree.sigouts[0].on(1)
        self.uhfqa.nodetree.sigouts[1].on(1)

        self.uhfqa.nodetree.awg.outputs[1].amplitude(-1.0)
        self.uhfqa.arm(length=self.repetitions, averages=1)
        # Configure output
        base, digital = self._get_shfsg_frequencies()
        out_channel = self.shfsg.nodetree.sgchannels[7]
        synth = self.shfsg.nodetree.synthesizers[out_channel.synthesizer()]

        out_channel.output.range(self.output_power)
        synth.centerfreq(base)
        out_channel.awg.modulation.enable(1)
        out_channel.oscs[0].freq(digital)

        self.uhfqa.compile_program(
            sequence_type=SequenceType.PULSED_SPEC,
            period=self.period,
            repetitions=self.repetitions,
            trigger_mode=TriggerMode.NONE,
            alignment=Alignment.START_WITH_TRIGGER
        )
        
        self.shfsg.nodetree.sgchannels[7].output.on(1)

        self.uhfqa.nodetree.qa.result.source(2)
        self.uhfqa.nodetree.sigouts[0].on(1)# Enable the outputs
        self.uhfqa.nodetree.sigouts[1].on(1)
        self.uhfqa.nodetree.awg.outputs[1].amplitude(-1.0)

        self._i_q_data = np.zeros(shape=(len(self.get_freqpoints()), 2), dtype=complex)

    @QkitFunction
    def start_measurement(self):
        print("Run")
        # Configure upper side band modulation
        out_channel = self.shfsg.nodetree.sgchannels[7]
        out_channel.sine.oscselect(0)
        out_channel.sine.i.enable(True)
        out_channel.sine.i.sin.amplitude(0.0)
        out_channel.sine.i.cos.amplitude(0.9 * self.amplitude_factor)
        out_channel.sine.q.enable(True)
        out_channel.sine.q.sin.amplitude(0.9 * self.amplitude_factor)
        out_channel.sine.q.cos.amplitude(0.0)
        self.uhfqa.arm(length=self.repetitions, averages=1)

        uhfqa_frequencies = self.get_freqpoints() - (self.get_freqpoints()[0] - MINIMUM_UHFQA_FREQUENCY)

        for i, f in enumerate(uhfqa_frequencies):
            self.uhfqa.nodetree.osc.freq(f)  # set modulation frequency
            for j in (0, 1):
                self.uhfqa.set_readout_frequency(f, j)
            
            self.uhfqa.run()                                    # start readout AWG                            # wait until trigger AWG has finished
            data = self.uhfqa.get_qubit_result()
            i_avg_result = np.mean(data[0]) # average the result vector
            q_avg_result = np.mean(data[1])
            self._i_q_data[i, :] = [i_avg_result, q_avg_result]         # append to results

    @QkitFunction
    def ready(self) -> bool:
        return True

    @QkitFunction
    def post_measurement(self):
        self.uhfqa.nodetree.sigouts[0].on(0)
        self.uhfqa.nodetree.sigouts[1].on(0)
        self.shfsg.nodetree.sgchannels[7].output.on(0)

    @QkitFunction
    def get_sweeptime(self, query=True):
        return 90

    @QkitFunction
    def get_sweeptime_averages(self):
        return 90
        


# server for providing TwoPortSwitch_RT by JB@KIT 04/2015 jochen.braumueller@kit.edu
# two port switch at room temperature to in situ (re)calibrate the qubit manipulation IQ-mixer
# model: Raspberry Pi running Raspian, Radiall Switch R572.432.000 (latching, no 50Ohms termination at open port) integrated in rack slot
# use: run on Raspberry Pi mounted in rack

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# install:
'''
    sudo apt-get update
    sudo apt-get install python-dev
    sudo apt-get install python-rpi.gpio
'''

'''
import RPi.GPIO as GPIO
GPIO.setmode(GPIO.BOARD)   #use RPi.GPIO layout

SWITCH1 = 16   #gpio23
SWITCH2 = 18   #gpio24
GPIO.setup(SWITCH1, GPIO.OUT)
GPIO.setup(SWITCH2, GPIO.OUT)   #set as outputs
GPIO.output(SWITCH1, GPIO.LOW)
GPIO.output(SWITCH2, GPIO.LOW)   #set off
'''

switch_position = 0

try:
    import SocketServer
except ImportError:
    import socketserver as SocketServer
from threading import Thread

#import os,sys
from time import sleep


class TCPHandler_Switch2port(SocketServer.BaseRequestHandler):
    """
    The RequestHandler class for our server.

    It is instantiated once per connection to the server, and must
    override the handle() method to implement communication to the
    client.
    """

    def switch(self, p):
        if p == 1:
            GPIO.output(SWITCH1, GPIO.HIGH)
            sleep(0.3)
            GPIO.output(SWITCH1, GPIO.LOW)
            switch_position = 1
            return True
        elif p == 2:
            GPIO.output(SWITCH2, GPIO.HIGH)
            sleep(0.3)
            GPIO.output(SWITCH2, GPIO.LOW)
            switch_position = 2
            return True
        else:
            return False
            
    def handle(self):
        try:
            # self.request is the TCP socket connected to the client
            self.data = self.request.recv(1024).strip()
            if str(self.data) == 'get_position':
                print 'switch get position request from', str(format(self.client_address[0]))
                self.request.sendall(str(switch_position))
            elif str(self.data) == 'switch 1':
                print 'switch 1 request from', str(format(self.client_address[0]))
                self.request.sendall(str(self.switch(1)))
            elif str(self.data) == 'switch 2':
                print 'switch 1 request from', str(format(self.client_address[0]))
                self.request.sendall(str(self.switch(2)))
            else:
                print 'request string not recognized'
                self.request.sendall(0)
        except Exception as m:
            print 'Error in TCP server handler:', m


class ThreadedTCPServer(SocketServer.ThreadingMixIn, SocketServer.TCPServer):   #enable multiple access
    pass


if __name__ == "__main__":
    HOST, PORT = 'ip-address', 9988

    # Create the server, binding to localhost on port 9955
    server = ThreadedTCPServer((HOST, PORT), TCPHandler_Switch2port)

    # Activate the server; this will keep running until you
    # interrupt the program with Ctrl-C
    try:
        server.serve_forever()
    finally:
        print 'shut down server'
        server.shutdown()




# zmq based info_service 
# HR@KIT/2017,2018


import zmq
import time
import logging
import qkit
from qkit.core.lib.com.signals import SIGNALS


class info_service(object):
    def __init__(self,host = None, port = None):
        if host:
            self.host = host
        else:
            self.host = qkit.cfg.get('info_host','localhost')
            
        if port:
            self.port = port
        else:
            self.port = qkit.cfg.get('info_port',5600)
            
        self.start_service(self.host,self.port)
    def start_service(self,host,port):
        socket_bound = False
        for newport in range(port,port+20):
            self.context = zmq.Context()
            self.socket = self.context.socket(zmq.PUB)
            try:
                self.socket.bind("tcp://*:%s" % newport)
                qkit.cfg['info_port'] = newport
                qkit.cfg['info_host'] = host
                socket_bound = True
                break
            except zmq.ZMQError:
                self.context.destroy()
        if not socket_bound:
            logging.warning("Info service: address/port in use: ZMQError. \nMaybe more than 20 instances of QKIT are running?")
            logging.warning("Not starting info service.")
            
        #zmq.ZMQError.errno
        # wait until zmq is settled
        time.sleep(0.3)
    
        
    def dist(self,topic, message = ""):
        "distribute a message"
        if SIGNALS.has_key(topic):
            sig = SIGNALS.get(topic)
            self.socket.send("%d:%s" % (sig, message))
        else:
            print("please specify a correct topic, specs are in SIGNALS.py")

    def info(self,message):
        self.dist('info',message)
        
    def emit(self,topic,message):
        "distribute a message to a topic(), message can be omitted"
        self.dist(topic,message)
        
    

if __name__ == "__main__":
    import sys
    if len(sys.argv) > 1:
        msg =  sys.argv[1]
    else: msg = ""
    
    ifs = info_service()
    
    ifs.info(msg)
    ifs.dist('reload',"quick!!! "+msg)
    



#!/usr/bin/env python
# srv_thread.py

# RP(Y)C factory object  for QKIT, written by HR,JB@KIT 2016

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


import time
import sys
import logging
#import numpy

import rpyc
from rpyc.utils.server import ThreadedServer

DEBUG = False 
def logstr(logstring):
    if DEBUG:
        print(str(logstring))

#global wants_abort
wants_abort = False

# the following trick allows to pass an Object (DATA) to a factory class (rpc.Service)
# witout overriding the __init__() function. Basically a closure defines the namespace
# and passes the object.
 
def CustomizedService(DATA):
    class Plasma1Service(rpyc.Service):
    
        def on_connect(self):
            # code that runs when a connection is created
            print 'New connection with %s initiated.'%str(self._conn._config['endpoints'][1])
    
        def on_disconnect(self):
            # code that runs when the connection has already closed
            print 'Connection to %s closed.'%str(self._conn._config['endpoints'][1])
            
        def exposed_get_history(self,p,range):   #range in hours
            if DATA.debug: print 'history request from client %s'%str(self._conn._config['endpoints'][1])
            return getattr(DATA,str(p)).get_history(range)
            
        def exposed_get_last_value(self,p):
            if DATA.debug: print 'value request'
            return [getattr(DATA,str(p)).get_timestamp(),getattr(DATA,str(p)).get_last_value()]
            
    return Plasma1Service

def remote_service_operator(DATA):
        """
        starts the (threaded) service which exposes the 
        RPYC_service factory class to remote RPYC clients.
            
        This is the function which should be called in the main function,
        there is no need for hiding it in a class.
        """
        CS  = CustomizedService(DATA)
        t = ThreadedServer(CS, port = DATA.localhost.port)#, protocol_config = {"allow_public_attrs" : True})
        t.start()

class MyTestObject(object):
    def __init__(self):
        self.a = 'string'
        
if __name__ == "__main__":
    """
    logic of operation:
    * init DATA
    * start remote service operator
    * start data_srv 
    """
    o = MyTestObject()
    o.test=lambda: 1
    rso = remote_service_operator(o)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: hannes.rotzinger@kit.edu / 2015,2016,2017
         marco.pfirrmann@kit.edu / 2016, 2017
@license: GPL
"""

import platform
if platform.system() == "Windows":
    try:
        import ctypes
        ctypes.windll.shcore.SetProcessDpiAwareness(False) # Fix Bug on Windows when using multiple screens with different scaling
    except:
        pass
import sys
# support both PyQt4 and 5
in_pyqt5 = False
in_pyqt4 = False
try:
    from PyQt5.QtCore import Qt, QObject
    from PyQt5.QtWidgets import QApplication
    in_pyqt5 = True
except ImportError as e:
    print("import of PyQt5 failed, trying PyQt4 (error: {})".format(e))

if not in_pyqt5:
    try:
        from PyQt4.QtCore import Qt, QObject
        from PyQt4.QtGui import QApplication
        in_pyqt4 = True
    except ImportError:
        print("import of PyQt5 and PyQt4 failed. Install one of those.")
        sys.exit(-1)


import argparse
import qkit
from qkit.gui.qviewkit.PlotWindow import PlotWindow
from threading import Lock

class DATA(QObject):
    """Class for housekeeping the different windows in qviewkit.

    This class handles the relationship; opening and closing of different
    PlotWindow objects.
    """

    open_plots = {}
    open_ds    = {}
    dataset_info = {}
    ds_tree_items= {}
    ds_cmd_open = {}
    toBe_deleted = []
    lock = Lock()
    info_thread_continue = True
    "a set of housekeeping functions..."

    def append_plot(self,parent,window_id,ds):
        # self, item, dataset

        window = PlotWindow(parent,self,ds)

        #print(window, window_id)
        self.open_plots[str(window_id)]=window
        self.open_ds[ds] = window_id

        window.show()   # non-modal

        return window
    def _toBe_deleted(self,ds):
        if ds in self.open_ds:
            self.toBe_deleted.append(ds)

    def _remove_plot_widgets(self, closeAll = False):
        # helper func to close widgets
        def close_ds(ds):
            if ds in self.open_ds:
                window_id = self.open_ds[ds]
                window_id.setCheckState(0,Qt.Unchecked)

                # make sure data is consitent
                if str(window_id) in self.open_plots:
                    self.open_plots[str(window_id)].deleteLater()
                    self.open_plots.pop(str(window_id))


                if ds in self.open_ds:
                    self.open_ds.pop(ds)

        if closeAll:
            for ds in list(self.open_ds.keys()):
                close_ds(ds)
        else:
            for ds in self.toBe_deleted:
                close_ds(ds)

        self.toBe_deleted = []

    def remove_plot(self,window_id,ds):

        if window_id in self.open_plots:
            win = self.open_plots.pop(window_id)
            win.deleteLater()

        if ds in self.open_ds:
            self.open_ds.pop(ds)


    def plot_is_open(self,ds):
        return ds in self.open_ds

    def has_dataset(self,ds):
        return ds in self.dataset_info

    def set_info_thread_continue(self,On):
        with self.lock:
            self.info_thread_continue = On
    def close_all(self):
        QApplication.quit()


# Main entry to program.
def main(argv=sys.argv):
    """Main function to open a h5 file with qviewkit.

    This function is called via command line. It opens the optional parsed
    h5 file and the optional parsed datasets automatically.
    The DatasetsWindow object creates a Qt based window with some UI buttons
    for file handling and a populated tree of the datasets.

    Args:
        '-f','--file',     type=str, help='hdf filename to open'
        '-ds','--datasets', type=str, help='(optional) datasets opened by default'

        '-rt','--refresh_time', type=float, help='(optional) refresh time'
        '-sp','--save_plot',  default=False,action='store_true', help='(optional) save default plots'
        '-live','--live_plot',default=False,action='store_true', help='(optional) if set, plots are reloaded'
        '-qinfo','--qkit_info',default=False,action='store_true', help='(optional) if set, listen to qkit infos'
    """
    # some configuration boilerplate
    data = DATA()

    parser = argparse.ArgumentParser(
        description="Qviewkit / qkit tool to visualize qkit-hdf files // HR@KIT 2015")


    parser.add_argument('-f','--file',     type=str, help='hdf filename to open')
    parser.add_argument('-ds','--datasets', type=str, help='(optional) datasets opened by default')

    parser.add_argument('-rt','--refresh_time', type=float, help='(optional) refresh time')
    parser.add_argument('-sp','--save_plot',  default=False,action='store_true', help='(optional) save default plots')
    parser.add_argument('-live','--live_plot',default=False,action='store_true', help='(optional) if set, plots are reloaded')
    parser.add_argument('-qinfo','--qkit_info',default=False,action='store_true', help='(optional) if set, listen to qkit infos')
    args=parser.parse_args(args=argv[1:])
    data.args = args

    # create Qt application
    if in_pyqt5:
        app = QApplication(argv)
    if in_pyqt4:
        app = QApplication(argv,True)

    # if activated, start info thread
    if data.args.qkit_info:
        from qkit.gui.qviewkit.info_subsys import info_thread
        it = info_thread(data)
        it.start()


    # create main window
    from qkit.gui.qviewkit.DatasetsWindow import DatasetsWindow
    #
    dsw = DatasetsWindow(data)
    dsw.show()
    dsw.raise_()

    # Connect signal for app quit
    app.lastWindowClosed.connect(quit)
    app.exec_()

if __name__ == "__main__":
    main()


# zmq  qkit info client, runs in a separate thread
# HR@KIT/2017


import zmq
import qkit
from qkit.core.lib.com.signals import SIGNALS

from threading import Thread
import time,sys

from qkit.core.lib.com.info_client import info_client


class info_thread(Thread):
   def __init__(self, data):
      Thread.__init__(self)
      self.data = data
      self.ifc = info_client()
      self.ifc.listen_to_all()
      self.data.set_info_thread_continue(True)
      
   def run(self):
      while self.data.info_thread_continue:
        socks = dict(self.ifc.poller.poll(1000))# wait 1s
        if self.ifc.socket in socks and socks[self.ifc.socket] == zmq.POLLIN:
            tid, message = self.ifc.get_message()
            print(message,tid)
            if int(tid)  == SIGNALS.get('close-gui'):
                print("received close gui signal")
                self.data.set_info_thread_continue(False)
                self.data.close_all()

        time.sleep(0.01)



if __name__ == "__main__":
    if len(sys.argv) > 1:
        host =  int(sys.argv[1])
        port =  int(sys.argv[2])
        ifc = info_client(host,port)
    else:
        ifc = info_client()
    ifc.listen_to_all()
    for i in range(10):
        ifc.print_message()



# instruments.py, code for a collection of instruments.
# Reinier heeres, <reinier@heeres.eu>, 2008
# HR @ KIT 2017

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
import qkit

import inspect
import os
import logging
import sys
import tomli
import qkit.core.instrument_base as instrument

import importlib


from qkit.core.lib.misc import get_traceback
TB = get_traceback()()

def _get_driver_module(name, do_reload=False):

    if name in sys.modules and not do_reload:
        return sys.modules[name]

    try:
        mod = importlib.import_module(name)
        if do_reload:
            importlib.reload(mod)
    except ImportError as e:
        fields = str(e).split(' ')
        if len(fields) > 0 and fields[-1] == name:
            logging.warning('Instrument driver %s not available', name)
        else:
            TB()
        return None
    except Exception as e:
        TB()
        logging.error('Error loading instrument driver %s', name)
        return None

    if name in sys.modules:
        return sys.modules[name]

    return None


class Insttools(object):

    __id = 1
    def __init__(self):
        Insttools.__id += 1

        self._instruments = {}
        self._instruments_info = {}
        self._tags = []
        self._instdir       = qkit.cfg.get('instruments_dir')
        self._user_instdir  = qkit.cfg.get('user_instruments_dir')
        lookup_dict_location = qkit.cfg.get('device_lookup_config')
        if lookup_dict_location is not None:
            self._lookup_dict = tomli.load(lookup_dict_location)
        else:
            self._lookup_dict = {}


    def __getitem__(self, key):
        return self.get(key)

    def __repr__(self):
        s = "Instruments list %s" % str(self.get_instrument_names())
        return s

    def add(self, ins, create_args={}):
        '''
        Add instrument to the internal instruments list and listen
        to signals emitted by the instrument.

        Input:  Instrument object
        Output: None
        '''

        self._instruments[ins.get_name()] = ins

        info = {'create_args': create_args}
        self._instruments_info[ins.get_name()] = info

        newtags = []
        for tag in ins.get_tags():
            if tag not in self._tags:
                self._tags.append(tag)
                newtags.append(tag)

    def get(self, name):
        '''
        Return Instrument object with name 'name'.

        Input:  name of instrument (string)
        Output: Instrument object
        '''

        if isinstance(name, instrument.Instrument):
            return name

        if type(name) == tuple:
            if len(name) != 1:
                return None
            name = name[0]

        if name in self._instruments:
            return self._instruments[name]
        else:
            return None

    def get_instrument_names(self):
        return sorted(self._instruments.keys())

    def get_instruments(self):
        '''
        Return the instruments dictionary of name -> Instrument.
        '''
        return self._instruments

    def get_types(self):
        '''
        Return list of supported instrument types
        '''
        ret = []
        filelist = os.listdir(self._instdir)
        for path_fn in filelist:
            path, fn = os.path.split(path_fn)
            name, ext = os.path.splitext(fn)
            if ext == '.py' and name[0] != '_':
                ret.append(name)
                
        if self._user_instdir:
            filelist = os.listdir(self._user_instdir)
            for path_fn in filelist:
                path, fn = os.path.split(path_fn)
                name, ext = os.path.splitext(fn)
                if ext == '.py' and name[0] != '_' and not ret.count(name) > 0:
                    ret.append(name)
        ret.sort()
        return ret

    def type_exists(self, typename):
        driverfn = os.path.join(self._instdir, '%s.py' % typename)
        if os.path.exists(driverfn):
            return True
        if self._user_instdir is None:
            return False
        driverfn = os.path.join(self._user_instdir, '%s.py' % typename)
        return os.path.exists(driverfn)
        
    def get_type_arguments(self, typename):
        '''
        Return info about the arguments of the constructor of 'typename'.

        Input:
            typename (string)
        Output:
            Tuple of (args, varargs, varkw, defaults)
            args: argument names
            varargs: name of '*' argument
            varkw: name of '**' argument
            defaults: default values
        '''

        module = _get_driver_module(typename)
        insclass = getattr(module, typename, None)
        if insclass is None:
            return None
        sig = inspect.signature(insclass.__init__)
        return list(sig.parameters.keys())

    def get_instruments_by_type(self, typename):
        '''
        Return all existing Instrument instances of type 'typename'.
        '''

        ret = []
        for name, ins in self._instruments.items():
            if ins.get_type() == typename:
                ret.append(ins)
        return ret

    def get_tags(self):
        '''
        Return list of tags present in instruments.
        '''

        return self._tags

    def _create_invalid_ins(self, name, instype, **kwargs):
        ins = instrument.InvalidInstrument(name, instype, **kwargs)
        self.add(ins, create_args=kwargs)
        return self.get(name)

    def create(self, name, instype, **kwargs):
        '''
        Create an instrument called 'name' of type 'type'.

        Input:  (1) name of the newly created instrument (string)
                (2) type of instrument (string)
                (3) optional: keyword arguments.
                    (1) tags, array of strings representing tags
                    (2) many instruments require address=<address>

        Output: Instrument object (Proxy)
        '''

        if not self.type_exists(instype):
            logging.error('Instrument type %s not supported', instype)
            return None

        if name in self._instruments:
            logging.warning('Instrument "%s" already exists, removing', name)
            self.remove(name)

        module = _get_driver_module(instype)
        if module is None:
            return self._create_invalid_ins(name, instype, **kwargs)
            
        importlib.reload(module)

        insclass = getattr(module, instype, None)
        if insclass is None:
            logging.error('Driver does not contain instrument class')
            return self._create_invalid_ins(name, instype, **kwargs)

        try:
            ins = insclass(name, **kwargs)
            for param_name in ins.get_parameter_names():
                if ins.get_parameter_options(param_name)['flags'] & ins.FLAG_GET: # do not query non-get or softget parameters
                    ins.get(param_name)  # Get all device parameters. This ensures that all get functions are working.
        except Exception as e:
            TB()
            logging.error('Error creating instrument %s: %s', name,e)
            return self._create_invalid_ins(name, instype, **kwargs)

        self.add(ins, create_args=kwargs)
        
        # Create a file where all created instruments with all parameters are stored once
        try:
            idn = ins.ask("*IDN?").strip()
        except:
            idn = "__none__"
        descr = str(name)+"#"+str(instype)+"#"+idn+"#"+str(kwargs)+"\r\n"
        fname = os.path.join(qkit.cfg['datadir'], "instrument.txt")  # save to datadir, because this is synced to backup server
        open(fname, "a").close() #create file if not existing
        with open(fname, "r+") as f:
            if not descr.strip() in [r.strip() for r in f.readlines()]:
                f.write(descr)
        return self.get(name)

    def create_from_config(self, name, **kwargs):
        """
        Check in the local config file for the device name. If it exists, take the config from there to create the device.
        
        Override with values from keyword arguments.
        """
        # This merges the configuration dict with
        config = {**self._lookup_dict[name], **kwargs, 'name': name}
        return self.create(**config)


    def reload_module(self, instype):
        module = _get_driver_module(instype, do_reload=True)
        return module is not None

    def reload(self, ins):
        '''
        Try to reload the module associated with instrument 'ins' and return
        the new instrument.

        In general about reloading: your milage may vary!

        Input:
            ins (Instrument or string): the instrument to reload

        Output:
            Reloaded instrument (Proxy)
        '''

        if type(ins) is bytes:
            ins = self.get(ins)
        if ins is None:
            return None

        insname = ins.get_name()
        instype = ins.get_type()
        kwargs = self._instruments_info[insname]['create_args']

        logging.info('reloading %r, type: %r, kwargs: %r',
                insname, instype, kwargs)

        self.remove(insname)
        reload_ok = self.reload_module(instype)
        if not reload_ok:
            return self._create_invalid_ins(insname, instype, **kwargs)

        return self.create(insname, instype, **kwargs)

    def auto_load(self, driver):
        '''
        Automatically load all instruments detected by 'driver' (an
        instrument_plugin module). This works only if it is supported by the
        driver by implementing a detect_instruments() function.
        '''

        module = _get_driver_module(driver)
        if module is None:
            return False
        importlib.reload(module)

        if not hasattr(module, 'detect_instruments'):
            logging.warning('Driver does not support instrument detection')
            return False

        devs = self.get_instruments_by_type(driver)
        for dev in devs:
            dev.remove()

        try:
            module.detect_instruments()
            return True
        except:
            return False

    def remove(self, name):
        '''
        Remove instrument from list and emit instrument-removed signal.

        Input:  (1) instrument name
        Output: None
        '''
        if name in self._instruments:
            del self._instruments[name]
            del self._instruments_info[name]

    class InstrumentBoundsError(ValueError):
        "Base Error to raise when instrument is out of bounds"
        pass

if __name__ == '__main__':
    i = Insttools()
    i.create('test1', 'HP1234')



# IQ cloud analysis
# started by M. Spiecker, 06/2020
#
# This IQ cloud analysis class performs a gaussian mixture fit on the IQ data to directly map out
# the density matrix of the system under study. The fit uses the sci-kit learn library.
#
# HOW TO USE IT:
# Either use the qubit cloud analysis routine
#
# or for more elaborate fits
#
# 0. Create an instance of IQCloudAnalysis
# 1. Load the I, Q data - (averages, sweep parameter)
# 2. Set up a gaussian mixture model.
# 4. Fit the clouds:
# 4.1 If the position and weights of the clouds are unkonwn you have to use the fit_clouds routine
#    The states are unsortet and have to be sortet either by weight or position.
# 4.2 For a qubit, when only the weights of the clouds are unkown, you can use the fast fit_clouds_weight function
# 5. Plot the gaussian mixture model on a given axes.
#
# Check out the notebook for examples!


import numpy as np
from scipy.optimize import fsolve
from scipy import constants as pyc
from sklearn.mixture import GaussianMixture

from matplotlib import colors


class IQCloudAnalysis:

    def __init__(self):

        self.sweep_param = None
        self.data = None
        self.dist_avg = 0.0
        self.n_steps = 0

        # fit
        self.gmm = None
        self.covariance_type=None
        self.n_states = 0
        self.positions = None
        self.weights = None
        self.covariances = None
        self.precisions = None
        self.max_variance = None
        self.generalized_variance = None

        # After sorting
        self.theta = 0.0
        self.angle = None
        self.separation = None
        self.beta = None
        self.temp = None

        self.amplitude = None
        self.phase = None

        # plots
        self.fig = None
        self.ax = None


    #######  out of the box usage for the two qubit clouds ################

    def analyse_qubit_clouds(self, I, Q, fq=-1,
                             pos_init=None, weights_init=None, precisions_init=None):
        """
        This routine performs an IQ cloud analyis with the given gaussian mixture model and
        interprets the two most prominent states as the qubit system.
        The rotation, the angle and the separation of the qubit clouds is determined.
        When the qubit frequency is given the temperature is calculated.
        """

        # 1. load data
        I = I[:, np.newaxis]  # extend the sweep axis
        Q = Q[:, np.newaxis]
        self.set_IQ_data_sweep(I, Q)

        # 2. set up gaussian micture model
        self.set_up_gaussian_mixture_model(self.n_states, covariance_type=self.covariance_type,
                                           pos_init=pos_init, weights_init=weights_init, precisions_init=precisions_init)

        # 3. fit clouds
        self.fit_clouds()

        # 4. sort clouds
        self.sort_clouds_weights()

        # analyze qubit clouds
        step = 0

        # calc theta
        di = self.positions[step, 0, 0] - self.positions[step, 1, 0]
        dq = self.positions[step, 0, 1] - self.positions[step, 1, 1]
        distance = np.sqrt(di**2 + dq**2)
        self.theta = np.sign(dq) * np.arccos(di / distance)

        # calc separation and angle
        a = np.linalg.norm(self.positions[step, 0, :])
        b = np.linalg.norm(self.positions[step, 1, :])
        self.separation = distance / np.sum(np.sqrt(self.generalized_variance[0,:2]) )  # lower bound for the separation
        self.angle = np.arccos((a**2 + b**2 - distance**2) / (2 * a * b))

        # calc temperature in mK
        if fq > 0:
            self.temp = 1e3 * pyc.h * fq / (pyc.k * np.log(self.weights[step, 0] / self.weights[step, 1]))
        else:
            self.temp = -1


    ################ load and manipulate data  ##############

    def set_IQ_data_sweep(self, I, Q, sweep_param=None, dist_avg=-1):
        """ Load 2-dimensional IQ data in the form [avg, sweep]
            If needed scale the data with a given dist_avg
        """

        data = np.dstack((I, Q))
        self.data = np.moveaxis(data, 1, 0)
        self.n_steps = self.data.shape[0]

        if sweep_param is None:
            self.sweep_param = np.arange(self.n_steps)
        else:
            self.sweep_param = sweep_param

        # scale data around unity - important for convergence
        if dist_avg == -1:
            self.dist_avg = np.mean(np.sqrt(self.data[:, :, 0]**2 + self.data[:, :, 1]**2), axis=(0, 1))
        else:
            self.dist_avg = dist_avg
        self.data /= self.dist_avg


    def rotate_data(self, theta):

        data = np.empty_like(self.data)
        data[:, :, 0] = np.cos(theta) * self.data[:, :, 0] - np.sin(theta) * self.data[:, :, 1]
        data[:, :, 1] = np.sin(theta) * self.data[:, :, 0] + np.cos(theta) * self.data[:, :, 1]
        self.data = data


    ######## gaussian mixture routines ############

    def set_up_gaussian_mixture_model(self, n_states=2, covariance_type='spherical',
                                      pos_init=None, weights_init=None, precisions_init=None, tol=1e-3):

        self.n_states = n_states
        self.covariance_type = covariance_type
        self.gmm = GaussianMixture(self.n_states, covariance_type=covariance_type,
                                   means_init=pos_init, weights_init=weights_init, precisions_init=precisions_init, tol=tol)

    def calc_covariance_properties(self, step):

        generalized_variance = np.empty(self.n_states)

        if self.covariance_type == "spherical":
            max_variance = np.max(self.covariances[step])
            generalized_variance[:] = self.covariances[step]
        elif self.covariance_type == "tied":
            e = np.linalg.eigvalsh(self.covariances[step])
            max_variance = np.max(e)
            generalized_variance[:] = np.linalg.det(self.covariances[step])
        elif self.covariance_type == "diag":
            max_variance = np.max(self.covariances[step])
            generalized_variance[:] = self.covariances[step][:, 0]**2 + self.covariances[step][:, 1]**2
        elif self.covariance_type == "full":
            max_variance = 0.0
            for i in range(self.n_states):
                e = np.linalg.eigvalsh(self.covariances[step][i, :, :])
                if np.max(e) > max_variance:
                    max_variance = np.max(e)
                generalized_variance[i] = np.linalg.det(self.covariances[step][i, :, :])

        return generalized_variance, max_variance

    def fit_clouds(self):
        """
        Fit of the clouds with the given gaussian mixture model.
        The clouds are randomly numerated and have to be sorted later on
        TODO: include option to use warm start of GaussianMixture to speed up convergence (using previous fit result)
        TODO: it would be nice if one could also set the inits in this routine: means_init=None, weights_init=None, precisions_init=None
        """
        self.positions = np.zeros((self.n_steps, self.n_states, 2))
        self.weights = np.zeros((self.n_steps, self.n_states))
        self.covariances = np.zeros(self.n_steps, dtype=object)
        self.precisions = np.zeros(self.n_steps, dtype=object)
        self.generalized_variance = np.zeros((self.n_steps, self.n_states))
        self.max_variance = np.zeros(self.n_steps)
        self.separations = np.zeros(self.n_steps)

        for i in range(self.n_steps):

            self.gmm.fit(self.data[i])

            self.positions[i, :, :] = self.gmm.means_
            self.weights[i, :] = self.gmm.weights_
            self.covariances[i] = self.gmm.covariances_
            self.precisions[i] = self.gmm.precisions_
            self.generalized_variance[i], self.max_variance[i] = self.calc_covariance_properties(i)
            di = self.positions[i, 0, 0] - self.positions[i, 1, 0]
            dq = self.positions[i, 0, 1] - self.positions[i, 1, 1]
            distance = np.sqrt(di ** 2 + dq ** 2)
            self.separations[i] = distance / np.sum(np.sqrt(self.generalized_variance[i,:2]))


    def fit_clouds_weight(self, ref_pos, ref_covar):
        """ Maximum likelihood fit of the weights.
            Simple implementation only for 2 states possible.
            TODO: for more states a warm start could give similar results
        """

        self.positions = np.zeros((self.n_steps, self.n_states, 2))
        self.weights = np.zeros((self.n_steps, self.n_states))
        self.covariances = np.zeros(self.n_steps, dtype=object)
        self.precisions = np.zeros(self.n_steps, dtype=object)
        self.generalized_variance = np.zeros((self.n_steps, self.n_states))
        self.max_variance = np.zeros(self.n_steps)

        if self.covariance_type == "spherical":
            ref_precs = 1 / ref_covar
        elif self.covariance_type == "tied":
            ref_precs = np.linalg(ref_covar)
        elif self.covariance_type == "diag":
            ref_precs = 1 / ref_covar
        elif self.covariance_type == "full":
            ref_precs = np.empty_like(ref_covar)
            for i in range(self.n_states):
                ref_precs[i, :, :] = np.linalg.inv(ref_covar[i, :, :])

        for i in range(self.n_steps):

            self.positions[i, :, :] = ref_pos
            self.covariances[i] = ref_covar
            self.precisions[i] = ref_precs
            self.generalized_variance[i], self.max_variance[i] = self.calc_covariance_properties(i)

            if self.n_states == 1:
                self.weights[i, 0] = 1.0
            if self.n_states == 2:
                self.weights[i, :] = self.weights_fit(i)
            else:
                print("Fitting of specific parameters is not supported by scikit-learn."
                      "You can try fitting with a warm start. TODO: implement it!")


    def weights_fit(self, step):
        """ Fast maximum likelihood fit of the weights for two states.
            The position and the variance must be given
        """
        s1 = np.sqrt(self.generalized_variance[step, 0])
        s2 = np.sqrt(self.generalized_variance[step, 1])
        dist = 0.5 * self.calc_mahalanobis_dist(self.data[step], self.positions[step], self.precisions[step])

        e = s2 / s1 * np.exp(dist[:, 1] - dist[:, 0])
        c = (e + 1) / (e - 1)

        tol = 1e-6
        delta_left = -1.0
        delta_right = 1.0
        delta = 0.0

        while delta_right - delta_left > tol:

            df = np.sum(1 / (delta + c))

            if df > 0:
                delta_left = delta
                delta = (delta_right + delta_left) / 2
            else:
                delta_right = delta
                delta = (delta_right + delta_left) / 2

                            # final numerical error of state population better than 0.5 * tol
        return 0.5 * (1 + delta), 0.5 * (1 - delta)


    def calc_mahalanobis_dist(self, data, positions, precisions):
        """
        Returns the squared mahalanobis distance
        """

        dist = np.empty((data.shape[0], self.n_states))

        for i in range(self.n_states):

            v = data - positions[i]
            if self.covariance_type == 'spherical':
                dist[:, i] = precisions[i] * np.einsum('ij,ij->i', v, v)
            elif self.covariance_type == 'tied':
                dist[:, i] = np.einsum('ij,ji->i', v, np.einsum('kj,ij->ki', precisions, v))
            elif self.covariance_type == 'diag':
                dist[:, i] = np.einsum('ij,ij->i', v, np.einsum('j,ij->ij', precisions[i], v))
            elif self.covariance_type == 'full':
                dist[:, i] = np.einsum('ij,ji->i', v, np.einsum('kj,ij->ki', precisions[i], v))

        return dist


    ####### different cloud sorting routines  ################

    def sort_clouds_positions(self, ref_pos=None):
        """ Sorts the states according to a reference positions.
            If ref_pos is None the first position of the sweep is chosen as reference.
        """
        if ref_pos is None:
            ref_pos = self.positions[0, :, :]

        label = np.zeros(self.n_states, dtype=int)

        for i in range(self.n_steps):
            for j in range(self.n_states):
                dist = (self.positions[i, j, 0] - ref_pos[:, 0])**2 + \
                       (self.positions[i, j, 1] - ref_pos[:, 1])**2

                label[j] = np.argmin(dist)

            self.positions[i, :, :] = self.positions[i, label, :]
            self.weights[i, :] = self.weights[i, label]
            if self.covariance_type != "tied":
                self.covariances[i][:] = self.covariances[i][label]
                self.precisions[i][:] = self.precisions[i][label]
            self.generalized_variance[i, :] = self.generalized_variance[i, label]


    def sort_clouds_weights(self):
        """ Sorts the states according to their weight
        """
        for i in range(self.n_steps):

            aw = np.argsort(self.weights[i, :])[::-1]
            self.positions[i, :, :] = self.positions[i, aw, :]
            self.weights[i, :] = self.weights[i, aw]
            if self.covariance_type != "tied":
                self.covariances[i][:] = self.covariances[i][aw]
                self.precisions[i][:] = self.precisions[i][aw]
            self.generalized_variance[i, :] = self.generalized_variance[i, aw]


    ###### cloud evaluation ##########

    def calc_amplitude_phase(self):

        self.amplitude = np.linalg.norm(self.positions[:, :, :], axis=2)
        self.phase = np.arctan2(self.positions[:, :, 1], self.positions[:, :, 0])

    def calc_cloud_distance(self, cloud_a, cloud_b):

        return np.linalg.norm(self.positions[:, cloud_b, :] - self.positions[:, cloud_a, :], axis=1)


    def calc_temperature(self, cloud_a, cloud_b, delta_freq):
        """
        Calculates the temperature from the weights for the
        corresponding frequencies of the states which must be given in units of Hz.
        If freq.size = n_states - 1 the zero ground state energy is automatically added.
        Consequently, for two states freq can also just be a number i.e. the frequency of the excited state
        """

        beta = np.log(self.weights[:, cloud_a] / self.weights[:, cloud_b]) / delta_freq

        self.temp = 1e3 * pyc.h / (pyc.k * beta)        # in mK
        self.beta = pyc.k * beta / pyc.h                # in 1 / K


    ###################### plotting #############################

    def set_axes(self, ax):
        self.ax = ax

    def plot_clouds(self, step=0, style="histogram", bin_size=0.015):

        x_range = [np.floor(np.min(self.data[step, :, 0]) / bin_size) * bin_size, np.ceil(np.max(self.data[step, :, 0]) / bin_size) * bin_size]
        y_range = [np.floor(np.min(self.data[step, :, 1]) / bin_size) * bin_size, np.ceil(np.max(self.data[step, :, 1]) / bin_size) * bin_size]
        bins_x = np.arange(x_range[0] - bin_size / 2, x_range[1] + bin_size, bin_size)
        bins_y = np.arange(y_range[0] - bin_size / 2, y_range[1] + bin_size, bin_size)

        im = None

        if style == "scatter":
            if self.data.shape[1] < 1e6:
                self.ax.scatter(self.data[step, :, 0], self.data[step, :, 1], 0.1)
            else:
                style = "histogram"
                print("More than 1e6 points - falling back on histogram!")

        if style == "histogram":

            H, _, _ = np.histogram2d(self.data[step, :, 0], self.data[step, :, 1], bins=[bins_x, bins_y])

            hmax = np.max(H)
            H[H < 1] = hmax + 1   # trick to get a white color for all zero entries

            im = self.ax.imshow(H.T, origin="lower", extent=(bins_x[0], bins_x[-1], bins_y[0], bins_y[-1]), cmap="terrain",
                      norm=colors.LogNorm(vmin=1, vmax=hmax), interpolation="none")

        # contour plot
        x = np.linspace(bins_x[0], bins_x[-1], 101)
        y = np.linspace(bins_y[0], bins_y[-1], 101)

        X, Y = np.meshgrid(x, y)

        iq = np.dstack((X, Y))
        iq = np.reshape(iq, (y.size * x.size, 2))
        dist = self.calc_mahalanobis_dist(iq, self.positions[step, :, :], self.precisions[step])
        dist = np.reshape(dist, (y.size, x.size, self.n_states))

        Z = np.zeros_like(X)
        for i in range(self.n_states):
            Z += self.weights[step, i] * np.exp(- 0.5 * dist[:, :, i]) / (2 * np.pi * np.sqrt(self.generalized_variance[step, i]))
        self.ax.contour(X, Y, np.log(Z))

        # TODO plot sigma areas of individual clouds
        # sigma areas of the individual states
        #s = np.linspace(0, 2 * np.pi, 101)
        #for i in range(self.n_states):
        #    self.ax.plot(self.positions[step, i, 0] + np.dot(np.cos(s), self.positions[step, i, 0] + np.sin(s))

        # plot origin and position of states
        # self.ax.plot(0.0, 0.0, 'r.')
        for i in range(self.n_states):
            self.ax.plot(self.positions[step, i, 0], self.positions[step, i, 1], 'k.')

        self.ax.set_xlim(bins_x[0], bins_x[-1])
        self.ax.set_ylim(bins_y[0], bins_y[-1])
        self.ax.axis("equal")

        return im


    def plot_weights(self):

        for i in range(self.n_states):
            self.ax.plot(self.sweep_param, self.weights[:, i], '.')







